from vpython import *

N = 10

scene.title = "{} by {} by {} = {} rotating boxes".format(N,N,N,N**3)
scene.append_to_title("\nGPU performance: <scan id='fps'/>")
##sphere()
##scene.pause()

boxes = []

L = 6
scene.range = L
length = 0.6*L/N
height = 0.4*L/N

for x in range(N):
    for y in range(N):
        for z in range(N):
            b = box(color=vector(x/N,y/N,z/N), 
                pos=vector(L*(x/(N-1)-.5),L*(y/(N-1)-.5),L*(z/(N-1)-.5)),
                size=vector(length,height,length))
            #b.axis = length*vector(1,1,0)
            boxes.append(b)

t = 0
dt = 0.01
loops = 0
ctime = 0
start = clock()
N = 200

while True:
    if loops == N:
        s = '{:0.1f} millisecond computation per loop'.format(1000*ctime/loops)
        s += '\n{:3.0f} loops per second'.format(loops/(clock()-start))
        scene.caption = s
        ctime = 0
        loops = 0
        start = clock()
    rate(N)
    ct = clock()
    v = length*vector(cos(t), 0, -sin(t))
    for b in boxes:
        b.axis = v
##        b.rotate(angle=dt, axis=vector(0,1,0))
    t += dt
    ctime += clock()-ct
    loops += 1



from vpython import *
scene.width = 600
scene.height = 600
scene.background = color.gray(0.9)
scene.range = 1.3
s = 'Pixel-level transparency using "depth-peeling."\n'
s += "Note the correct transparencies of the intersecting slabs.\n"
scene.title = s

box(pos=vec(0,0,0), opacity=1, size=vec(1,1,1), texture=textures.flower)
sphere(pos=vector(0,0,.9), opacity=0.3, shininess=0, radius=0.2, color=color.green)
s = sphere(pos=vector(0.1,0,1.2), opacity=0.2, shininess=0, radius=0.1, color=color.cyan)
box(pos=s.pos, size=0.06*vector(1,1,1), color=color.gray(.2))
box(pos=vector(0,.5,1), color=color.red,  opacity=0.2, size=vector(.05,.2,.8), axis=vector(1,0,1) )
box(pos=vector(0,.5,1), color=color.cyan, opacity=0.2, size=vector(.05,.2,.8), axis=vector(1,0,-1))

scene.caption = """Right button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
     On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""



import sys
import vpython

API_NAMES = [
    'Camera',
    'GSprint',
    'GSversion',
    'GlowWidget',
    'Mouse',
    'RackOutline',
    'ToothOutline',
    'acos',
    'acosh',
    'addpos',
    'adjust_axis',
    'adjust_up',
    'arange',
    'arrow',
    'asin',
    'asinh',
    'atan',
    'atan2',
    'atanh',
    'attach_arrow',
    'attach_light',
    'attach_trail',
    'baseObj',
    'box',
    'bumpmaps',
    'button',
    'canvas',
    'ceil',
    'checkbox',
    'clock',
    'color',
    'combin',
    'comp',
    'compound',
    'cone',
    'controls',
    'convert',
    'copysign',
    'cos',
    'cosh',
    'cross',
    'curve',
    'curveMethods',
    'cylinder',
    'degrees',
    'diff_angle',
    'distant_light',
    'dot',
    'e',
    'ellipsoid',
    'erf',
    'erfc',
    'event_return',
    'exp',
    'expm1',
    'extrusion',
    'fabs',
    'faces',
    'factorial',
    'floor',
    'fmod',
    'frame',
    'frexp',
    'fsum',
    'gamma',
    'gcd',
    'gcurve',
    'gdots',
    'ghbars',
    'gobj',
    'graph',
    'gs_version',
    'gvbars',
    'hat',
    'helix',
    'hypot',
    'inf',
    'isclose',
    'isfinite',
    'isinf',
    'isnan',
    'keysdown',
    'label',
    'ldexp',
    'lgamma',
    'local_light',
    'log',
    'log10',
    'log1p',
    'log2',
    'mag',
    'mag2',
    'menu',
    'meta_canvas',
    'modf',
    'nan',
    'norm',
    'object_rotate',
    'path_object',
    'paths',
    'pi',
    'points',
    'pow',
    'proj',
    'pyramid',
    'quad',
    'radians',
    'radio',
    'random',
    'rate',
    'rate_control',
    'ring',
    'rotate',
    'rotatecp',
    'roundc',
    'scalecp',
    'scene',
    'set_browser',
    'shape_object',
    'shapes',
    'shapespaths',
    'simple_sphere',
    'sin',
    'sinh',
    'sleep',  # From vpython.py
    'slider',
    'sphere',
    'sqrt',
    'standardAttributes',
    'tan',
    'tanh',
    'tau',
    'text',
    'textures',
    'triangle',
    'trunc',
    'vec',
    'vector',
    'version',
    'vertex',
    'winput',
    'wtext'
]


def test_names_in_base_namspace():
    current_names = set(name for name in dir(vpython)
                        if not name.startswith('_'))
    api_name_set = set(API_NAMES)

    python_version = sys.version_info

    # Python 3.7 added remainder to math, so add it to what we
    # expect to see.
    if python_version.major == 3 and python_version.minor >= 7:
        api_name_set.add('remainder')

    # Python 3.8 added even more math functions.
    if python_version.major == 3 and python_version.minor >= 8:
        for name in ['dist', 'comb', 'prod', 'perm', 'isqrt']:
            api_name_set.add(name)

    # Python 3.9 adds three more new math functions.
    if python_version.major == 3 and python_version.minor >= 9:
        for name in ['lcm', 'ulp', 'nextafter']:
            api_name_set.add(name)

    print(sorted(api_name_set - current_names))

    # We may have added new names, so start with this weaker test
    assert api_name_set.issubset(current_names)

    # Ideally this test also passes -- everything in the API is what
    # we should expose and everything we expose is in the API.
    assert api_name_set == current_names



from vpython import *
scene.caption = """In GlowScript programs:
Right button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
  On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""
scene.forward = vector(0,-.3,-1)

G = 6.7e-11 # Newton gravitational constant

giant = sphere(pos=vector(-1e11,0,0), radius=2e10, color=color.red, 
                make_trail=True, trail_type='points', interval=10, retain=50)
giant.mass = 2e30
giant.p = vector(0, 0, -1e4) * giant.mass

dwarf = sphere(pos=vector(1.5e11,0,0), radius=1e10, color=color.yellow,
                make_trail=True, interval=10, retain=50)
dwarf.mass = 1e30
dwarf.p = -giant.p

dt = 1e5
while True:
    rate(200)
    r = dwarf.pos - giant.pos
    F = G * giant.mass * dwarf.mass * norm(r) / mag2(r)
    giant.p = giant.p + F*dt
    dwarf.p = dwarf.p - F*dt
    giant.pos = giant.pos + (giant.p/giant.mass) * dt
    dwarf.pos = dwarf.pos + (dwarf.p/dwarf.mass) * dt



from vpython import *
scene.caption = """Right! button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
     On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""

#scene.title = "<script src='https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.2/MathJax.js?config=TeX-MML-AM_CHTML'></script>"
#scene.title += "\\(\\dfrac {5} {7} \\)"
#print(scene.title)

side = 4.0
thk = 0.3
s2 = 2*side - thk
s3 = 2*side + thk

wallR = box (pos=vector( side, 0, 0), size=vector(thk, s2, s3),  color = color.red)
wallL = box (pos=vector(-side, 0, 0), size=vector(thk, s2, s3),  color = color.red)
wallB = box (pos=vector(0, -side, 0), size=vector(s3, thk, s3),  color = color.blue)
wallT = box (pos=vector(0,  side, 0), size=vector(s3, thk, s3),  color = color.blue)
wallBK = box(pos=vector(0, 0, -side), size=vector(s2, s2, thk), color = color.gray(0.7))

ball = sphere (pos=vec(1,1.5,0), color = color.green, radius = 0.4, make_trail=True, retain=200)
ball.mass = 1.0
ball.p = vector (-0.15, -0.23, +0.27)

side = side - thk*0.5 - ball.radius
dt = 0.3
while True:
    rate(200)
    ball.pos = ball.pos + (ball.p/ball.mass)*dt
    if not (side > ball.pos.x > -side):
        ball.p.x = -ball.p.x
    if not (side > ball.pos.y > -side):
        ball.p.y = -ball.p.y
    if not (side > ball.pos.z > -side):
        ball.p.z = -ball.p.z




from vpython import *

scale = 4e-14/1e17
ec = 1.6e-19  # electron charge
kel = 9e9     # Coulomb constant
scene.range = 2e-13

charges = [ sphere( pos=vector(-1e-13,0,0), Q=ec,  color=color.red,  size=1.2e-14*vector(1,1,1) ),
            sphere( pos=vector( 1e-13,0,0), Q=-ec, color=color.blue, size=1.2e-14*vector(1,1,1) )]

s = "Click or drag to plot an electric field vector produced by the two charges.\n"
s += "On a touch screen, tap, or press and hold, then drag.\n"
s += "Arrows representing the field are bluer if low magnitude, redder if high."
scene.caption = s

def getfield(p):
    f = vec(0,0,0)
    for c in charges:
        f = f + (p-c.pos) * kel * c.Q / mag(p-c.pos)**3
    return f

def mouse_to_field(a):
    p = scene.mouse.pos
    f = getfield(p)
    m = mag(f)
    red = max( 1-1e17/m, 0 )
    blue = min(  1e17/m, 1 )
    if red >= blue:
        blue = blue/red
        red = 1.0
    else:
        red = red/blue
        blue = 1.0
    a.pos = p
    a.axis = scale*f
    a.color = vector(red,0,blue)
    a.visible = True

drag = False
a = None

def down(ev):
    global a, drag
    a = arrow(shaftwidth=6e-15, visible=False)
    mouse_to_field(a)
    drag = True

def move(ev):
    global a, drag
    if not drag: return
    mouse_to_field(a)
    
def up(ev):
    global a, drag
    mouse_to_field(a)
    drag = False

scene.bind("mousedown", down)

scene.bind("mousemove", move)

scene.bind("mouseup", up)



from vpython import *
import povexport
# This program displays most GlowScript 3D objects.

#It also illustrates key features such as mouse handling, rate, and sleep.

# Bruce Sherwood, August 2011; VPython version November 2014

scene.background = color.gray(0.7)
scene.center = vector(0,0.5,0)
scene.forward = vector(-.3,0,-1)
gslabel = label(pos=vector(1.5,2,0), text='VPython', color=color.blue, align='left', box=0, line=0)
box(pos=vector(-2,0,0), size=vector(.3,2.5,2.5), pov_texture='T_Wood24')
box(pos=vector(.25,-1.4,0), size=vector(4.8,.3,2.5), pov_texture='T_Wood24')
cylinder(pos=vector(-2,2,1.25), radius=0.7, axis=vector(0,0,-2.5), color=color.red)
ball = sphere(pos=vector(2,1,0), radius=0.5, color=color.cyan)
ptr = arrow(pos=vector(0,0,2), axis=vector(2,0,0), color=color.yellow)
cone(pos=vector(-2,0,0), radius=1, length=3, color=color.green, opacity=0.3)
ring(pos=vector(.2,0,0), radius=.6, axis=vector(1,0,0), thickness=0.12, color=color.gray(0.4))
ellipsoid(pos=vector(-.3,2,0), color=color.orange, size=vector(.3,1.5,1.5))
pyramid(pos=vector(.3,2,0), color=vector(0,0.5,.25), size=vector(0.8,1.2,1.2))
spring = helix(pos=vector(2,-1.25,0), radius=0.3, axis=vector(0,1.8,0),
        color=color.orange, thickness=.1)
angle = 0
da = .01

trail = curve(color=color.magenta, radius= .02)
trail.append(vector(1,0,0))
trail.append(vector(1,0,2))
trail.append(vector(2,0,2))

while angle < 3*pi/4:
  ptr.rotate(angle=da, axis=vector(0,0,1), origin=ptr.pos)
  trail.append(ptr.pos+ptr.axis)
  angle += da

inclist = ['colors.inc', 'stones.inc', 'woods.inc', 'metals.inc']
scene.pause('Adjust the camera, then click to export to POV-ray.')
povexport.export(canvas=scene, filename='VPobjects.pov', include_list=inclist)



import os
import time
from threading import Thread

from jupyter_core.paths import jupyter_data_dir
import notebook
import IPython
from IPython.display import display, Javascript

from .vpython import GlowWidget, baseObj, canvas
from .rate_control import ws_queue
from . import __version__

import tornado.httpserver
import tornado.websocket
import tornado.ioloop
import tornado.web
import socket
import json
import asyncio
import logging

def find_free_port():
    s = socket.socket()
    s.bind(('',0)) # find an available port
    return s.getsockname()[1]

__SOCKET_PORT = find_free_port()

try:
    if platform.python_implementation() == 'PyPy':
        __SOCKET_PORT = 9000 + __SOCKET_PORT % 1000     # use port number between 9000 and 9999 for PyPy
except:
    pass


#### Setup for Jupyter VPython

# The following file operations check whether nbextensions already has the correct files.
package_dir = os.path.dirname(__file__) # The location in site-packages of the vpython module
datacnt = len(os.listdir(package_dir+"/vpython_data"))     # the number of files in the site-packages vpython data folder
libcnt = len(os.listdir(package_dir+"/vpython_libraries")) # the number of files in the site-packages vpython libraries folder
jd = jupyter_data_dir()
nbdir = jd+'/nbextensions/'
nbdata = nbdir+'vpython_data'
nblib = nbdir+'vpython_libraries'
transfer = True # need to transfer files from site-packages to nbextensions

### If JupyterLab is installed then copy vpython_data directory to static dir in Jupytarlab Application Directory
try:
	import jupyterlab
	import jupyterlab.commands
except ImportError:
	#logging.info("Unable to import jupyterlab")
	pass
else:
    # We have jupyterlab, is it the right version?
    if (jupyterlab.__version__ >= '0.35.0') and (jupyterlab.__version__ < '1.0.0'):
        from os.path import join
        labextensions_dir = join(jupyterlab.commands.get_app_dir(), u'static')
        try:
            notebook.nbextensions.install_nbextension(path=package_dir + "/vpython_data",
                	                              nbextensions_dir=labextensions_dir,
                        	                      overwrite=False,
                                                      verbose=0)
        except PermissionError:
            #logging.info("PermissionError: Unable to install /vpython_data directory and files for VPython on JupyterLab")
            pass


if 'nbextensions' in os.listdir(jd):
    ldir = os.listdir(nbdir)
    if ('vpython_data' in ldir and len(os.listdir(nbdata)) == datacnt and
       'vpython_libraries' in ldir and len(os.listdir(nblib)) == libcnt and
        'vpython_version.txt' in ldir):
        v = open(nbdir+'/vpython_version.txt').read()
        transfer = (v != __version__) # need not transfer files to nbextensions if correct version's files already there

if transfer:
    notebook.nbextensions.install_nbextension(path = package_dir+"/vpython_data",overwrite = True,user = True,verbose = 0)
    notebook.nbextensions.install_nbextension(path = package_dir+"/vpython_libraries",overwrite = True,user = True,verbose = 0)

    # Wait for files to be transferred to nbextensions:
    libready = False
    dataready = False
    while True:
        nb = os.listdir(nbdir)
        for f in nb:
            if f == 'vpython_data':
                if len(os.listdir(nbdata)) == datacnt:
                    dataready = True
            if f == 'vpython_libraries':
                if len(os.listdir(nblib)) == libcnt:
                    libready = True
        if libready and dataready: break
    # Mark with the version number that the files have been transferred successfully:
    fd = open(nbdir+'/vpython_version.txt', 'w')
    fd.write(__version__)
    fd.close()

display(Javascript("""if (typeof Jupyter !== "undefined") {require.undef("nbextensions/vpython_libraries/glow.min");}else{element.textContent = ' ';}"""))
display(Javascript("""if (typeof Jupyter !== "undefined") {require.undef("nbextensions/vpython_libraries/glowcomm");}else{element.textContent = ' ';}"""))
display(Javascript("""if (typeof Jupyter !== "undefined") {require.undef("nbextensions/vpython_libraries/jquery-ui.custom.min");}else{element.textContent = ' ';}"""))

display(Javascript("""if (typeof Jupyter !== "undefined") {require(["nbextensions/vpython_libraries/glow.min"], function(){console.log("GLOW LOADED");});}else{element.textContent = ' ';}"""))
display(Javascript("""if (typeof Jupyter !== "undefined") {require(["nbextensions/vpython_libraries/glowcomm"], function(){console.log("GLOWCOMM LOADED");});}else{element.textContent = ' ';}"""))
display(Javascript("""if (typeof Jupyter !== "undefined") {require(["nbextensions/vpython_libraries/jquery-ui.custom.min"], function(){console.log("JQUERY LOADED");});}else{element.textContent = ' ';}"""))

if transfer:
    time.sleep(4)      # allow some time for javascript code above to run after nbextensions update before attempting to setup Comm Channel
else:
    time.sleep(2)      # allow some time for javascript code above to run before attempting to setup Comm Channel

wsConnected = False

class WSHandler(tornado.websocket.WebSocketHandler):
    def open(self):
        global wsConnected
        wsConnected = True

    def on_message(self, message):
        ws_queue.put(message)

    def on_close(self):
        self.stop_tornado()

    def stop_tornado(self):
        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.add_callback(ioloop.stop)

    def check_origin(self, origin):
        return True

def start_server():
    asyncio.set_event_loop(asyncio.new_event_loop())
    application = tornado.web.Application([(r'/ws', WSHandler),])
    http_server = tornado.httpserver.HTTPServer(application)
    http_server.listen(__SOCKET_PORT)
    Log = logging.getLogger('tornado.access')
    level = logging.getLevelName('WARN')
    Log.setLevel(level)
    tornado.ioloop.IOLoop.instance().start()


# Removed check for ipykernel version because the old check
# was for 5.0.0 but this works with 4.x too...and 4.x is the first
# version of ipykernel
t = Thread(target=start_server, args=())
t.start()
# Setup Comm Channel and websocket
baseObj.glow = GlowWidget(wsport=__SOCKET_PORT, wsuri='/ws')
while (not wsConnected):
    time.sleep(0.1)          # wait for websocket to connect

baseObj.trigger()  # start the trigger ping-pong process


# Same justification as above for removing the ipykernel check.
# There was also an IPython version check for >=4, which was
# released in Nov 2015. Ok to stop supporting in 2.019 I think.
async def wsperiodic():
    while True:
        if ws_queue.qsize() > 0:
            data = ws_queue.get()
            d = json.loads(data)
            # Must send events one at a time to GW.handle_msg because
            # bound events need the loop code
            for m in d:
                # message format used by notebook
                msg = {'content': {'data': [m]}}
                baseObj.glow.handle_msg(msg)

        await asyncio.sleep(0.1)

loop = asyncio.get_event_loop()
loop.create_task(wsperiodic())

# Dummy name to import...
_ = None



import sys
import PyQt5.QtCore
import PyQt5.QtWebEngineWidgets
from PyQt5.QtWidgets import QApplication


if len(sys.argv) > 1:

    if sys.argv[1]:

        app = QApplication(sys.argv)

        web = PyQt5.QtWebEngineWidgets.QWebEngineView()
        web.load(PyQt5.QtCore.QUrl(sys.argv[1]))
        web.show()

        sys.exit(app.exec_())

    else:
        print("Please give a URL as the first command-line argument "
              "when running the program.")

else:
    print("Please give a URL as the first command-line argument "
          "when running the program.")



from vpython import *
# This version uses VPython widgets: button, radio button, checkbox, slider, menu
# See ButtonsSlidersMenus1 for a version that uses Jupyter notebook widgets: button, slider, menu
scene.width = 350
scene.height = 300
scene.range = 1.3
scene.title = "Widgets (buttons, etc.)\n"

running = True

def Run(b):
    global running
    running = not running
    if running: b.text = "Pause"
    else: b.text = "Run"
    
button(text="Pause", pos=scene.title_anchor, bind=Run)

box_object = box(visible=True)
cone_object = cone(visible=False, radius=0.5)
pyramid_object = pyramid(visible=False)
cylinder_object = cylinder(visible=False, radius=0.5)
sphere(radius=0.3)

col = color.cyan
currentobject = box_object
currentobject.color = col

def Color(c):
    global col
    if col.equals(color.cyan): # change to red
        currentobject.color = col = color.red
        cbutton.text = "<b>Cyan</b>"
        cbutton.textcolor = color.cyan
        cbutton.background = color.red
        r1.checked = False
        r2.checked = True
    else:                      # change to cyan
        currentobject.color = col = color.cyan
        cbutton.text = "<b>Red</b>"
        cbutton.textcolor = color.red
        cbutton.background = color.cyan
        r1.checked = True
        r2.checked = False
        
cbutton = button(text='<b>Red</b>', textcolor=color.red, background=color.cyan, pos=scene.title_anchor, bind=Color)

scene.caption = "Vary the rotation speed:\n"
speed = 150
def setspeed(s):
    global speed
    speed = s.value
    wt.text = '{:1.0f}'.format(s.value)
    
sl = slider(min=20, max=500, value=250, length=280, bind=setspeed, right=15)

wt = wtext(text='{:1.0f}'.format(sl.value))

scene.append_to_caption('\n')

r1 = radio(bind=Color, checked=True, text='Cyan')

scene.append_to_caption('       ')

def M(m):
    global col, currentobject
    op = currentobject.opacity
    currentaxis = currentobject.axis
    currentobject.visible = False
    val = m.selected
    if val == "box": 
        currentobject = box_object
    elif val == "cone": 
        currentobject = cone_object
    elif val == "pyramid": 
        currentobject = pyramid_object
    elif val == "cylinder": 
        currentobject = cylinder_object
    currentobject.color = col
    currentobject.axis = currentaxis
    currentobject.visible = True
    currentobject.opacity = op

menu(choices=['Choose an object', 'box', 'cone', 'pyramid', 'cylinder'], index=0, bind=M)

scene.append_to_caption('\n')

r2 = radio(bind=Color, text='Red')

scene.append_to_caption('         ')

def transparency(b):
    if b.checked:
        currentobject.opacity = 0.5
    else:
        currentobject.opacity = 1

checkbox(bind=transparency, text='Transparent')

while True:
    rate(100)
    if running:
        currentobject.rotate(angle=speed*1e-4, axis=vector(0,1,0))




from vpython import *
# John Coady

title = "Click and drag the mouse in the 3D canvas to insert and drag a small sphere."
scene.title = title

scene.range = 3

b = box(pos=vector(0,0,0), color=color.green)

drag = True
s = None

def grab(evt):
    global s, drag
    scene.title = 'Drag the sphere.'
    drag = True
    s = sphere(pos=evt.pos, radius=0.1, color=color.red)

def move(evt):
    if drag:
        s.pos = scene.mouse.pos #evt.pos
        
def drop(evt):
    global drag
    scene.title = title
    s.color = color.cyan
    drag = False

scene.bind('mousedown',grab)
scene.bind('mousemove',move)
scene.bind('mouseup',drop)

while True:
    rate(1)



from vpython import *
# Double pendulum

# The analysis is in terms of Lagrangian mechanics.
# The Lagrangian variables are angle of upper bar, angle of lower bar,
# measured from the vertical.

# Bruce Sherwood

# Corrections to the Lagrangian calculations by Owen Long, UC. Riverside

scene.width = scene.height = 600
scene.range = 1.8
scene.title = "A double pendulum"

def display_instructions():
    s = "In VPython programs:\n"
    s += "    Rotate the camera by dragging with the right mouse button,\n        or hold down the Ctrl key and drag.\n"
    s += "    To zoom, drag with the left+right mouse buttons,\n         or hold down the Alt/Option key and drag,\n         or use the mouse wheel.\n"
    s += "Touch screen: pinch/extend to zoom, swipe or two-finger rotate."
    scene.caption = s

# Display text below the 3D graphics:
display_instructions()

g = 9.8
M1 = 2.0
M2 = 1.0
d = 0.05 # thickness of each bar
gap = 2*d # distance between two parts of upper, U-shaped assembly
L1 = 0.5 # physical length of upper assembly; distance between axles
L1display = L1+d # show upper assembly a bit longer than physical, to overlap axle
L2 = 1 # physical length of lower bar
L2display = L2+d/2 # show lower bar a bit longer than physical, to overlap axle
# Coefficients used in Lagrangian calculation
A = (1/4)*M1*L1**2+(1/12)*M1*L1**2+M2*L1**2
B = (1/2)*M2*L1*L2
C = g*L1*(M1/2+M2)
D = M2*L1*L2/2
E = (1/12)*M2*L2**2+(1/4)*M2*L2**2
F = g*L2*M2/2

hpedestal = 1.3*(L1+L2) # height of pedestal
wpedestal = 0.1 # width of pedestal
tbase = 0.05 # thickness of base
wbase = 8*gap # width of base
offset = 2*gap # from center of pedestal to center of U-shaped upper assembly
pedestal_top = vec(0,hpedestal/2,0) # top of inner bar of U-shaped upper assembly

theta1 = 1.3*pi/2 # initial upper angle (from vertical)
theta1dot = 0 # initial rate of change of theta1
theta2 = 0 # initial lower angle (from vertical)
theta2dot = 0 # initial rate of change of theta2

pedestal = box( pos=pedestal_top-vec(0,hpedestal/2,offset),
                size=vec(wpedestal,1.1*hpedestal,wpedestal),
                color=vec(0.4,0.4,0.5) )
base = box( pos=pedestal_top-vec(0,hpedestal+tbase/2,offset),
                 size=vec(wbase,tbase,wbase),
                 color=pedestal.color )
axle1 = cylinder( pos=pedestal_top-vec(0,0,gap/2-d/4), axis=vec(0,0,-1),
                 size=vec(offset,d/4,d/4), color=color.yellow )

bar1 = box( pos=pedestal_top+vec(L1display/2-d/2,0,-(gap+d)/2), 
                 size=vec(L1display,d,d), color=color.red )
bar1.rotate( angle=-pi/2, axis=vec(0,0,1), origin=vec(axle1.pos.x, axle1.pos.y, bar1.pos.z) )
bar1.rotate( angle=theta1, axis=vec(0,0,1), origin=vec(axle1.pos.x, axle1.pos.y, bar1.pos.z) )

bar1b = box( pos=pedestal_top+vec(L1display/2-d/2,0,(gap+d)/2), 
                 size=vec(L1display,d,d), color=bar1.color )
bar1b.rotate( angle=-pi/2, axis=vec(0,0,1), origin=vec(axle1.pos.x, axle1.pos.y, bar1b.pos.z) )
bar1b.rotate( angle=theta1, axis=vec(0,0,1), origin=vec(axle1.pos.x, axle1.pos.y, bar1b.pos.z) )

pivot1 = vec(axle1.pos.x, axle1.pos.y, 0)

axle2 = cylinder( pos=pedestal_top+vec(L1,0,-(gap+d)/2), axis=vec(0,0,1), 
                size=vec(gap+d,axle1.size.y/2,axle1.size.y/2), color=axle1.color )
axle2.rotate( angle=-pi/2, axis=vec(0,0,1), origin=vec(axle1.pos.x, axle1.pos.y, axle2.pos.z) )
axle2.rotate( angle=theta1, axis=vec(0,0,1), origin=vec(axle1.pos.x, axle1.pos.y, axle2.pos.z) )

bar2 = box( pos=axle2.pos+vec(L2display/2-d/2,0,(gap+d)/2), 
        size=vec(L2display,d,d), color=color.green )

bar2.rotate( angle=-pi/2,  axis=vec(0,0,1), origin=vec(axle2.pos.x, axle2.pos.y, bar2.pos.z) )
bar2.rotate( angle=theta2,  axis=vec(0,0,1), origin=vec(axle2.pos.x, axle2.pos.y, bar2.pos.z) )

dt = 0.001
t = 0

while True:
    rate(1/dt) 
    # Calculate accelerations of the Lagrangian coordinates=
    atheta1 = ((E*C/B)*sin(theta1)-F*sin(theta2))/(D-E*A/B)
    atheta2 = -(A*atheta1+C*sin(theta1))/B
    # Update velocities of the Lagrangian coordinates=
    theta1dot = theta1dot+atheta1*dt
    theta2dot = theta2dot+atheta2*dt
    # Update Lagrangian coordinates=
    dtheta1 = theta1dot*dt
    dtheta2 = theta2dot*dt
    theta1 = theta1+dtheta1
    theta2 = theta2+dtheta2
    
    bar1.rotate( angle=dtheta1, axis=vec(0,0,1), origin=pivot1 )
    bar1b.rotate( angle=dtheta1, axis=vec(0,0,1), origin=pivot1 )
    pivot2 = vec(axle2.pos.x, axle2.pos.y, pivot1.z)
    axle2.rotate( angle=dtheta1, axis=vec(0,0,1), origin=pivot1 )
    bar2.rotate( angle=dtheta2, axis=vec(0,0,1), origin=pivot2 )
    pivot2 = vec(axle2.pos.x, axle2.pos.y, pivot1.z)
    bar2.pos = pivot2 + bar2.axis/2
    
    t = t+dt



## vectors and associated methods
from math import cos, sin, acos, sqrt, pi
from random import random

# List of names imported from this module with import *
__all__ = ['adjust_axis', 'adjust_up', 'comp', 'cross', 'diff_angle', 'dot',
           'hat', 'mag', 'mag2', 'norm', 'object_rotate', 'proj', 'rotate',
           'vector']


class vector(object):
    'vector class'

    @staticmethod
    def random():
        return vector(-1.0 + 2.0*random(), -1.0 + 2.0*random(), -1.0 + 2.0*random())

    def __init__(self, *args):
        if len(args) == 3:
            self._x = float(args[0]) # make sure it's a float; could be numpy.float64
            self._y = float(args[1])
            self._z = float(args[2])
        elif len(args) == 1 and isinstance(args[0], vector): # make a copy of a vector
            other = args[0]
            self._x = other._x
            self._y = other._y
            self._z = other._z
        else:
            raise TypeError('A vector needs 3 components.')
        self.on_change = self.ignore

    def ignore(self):
        pass

    @property
    def value(self):
        return [self._x, self._y, self._z]
    @value.setter
    def value(self,other):  ## ensures a copy; other is a vector
        self._x = other._x
        self._y = other._y
        self._z = other._z

    def __neg__(self):
        return vector(-self._x, -self._y, -self._z)

    def __pos__(self):
        return self

    def __str__(self):
        return '<{:.6g}, {:.6g}, {:.6g}>'.format(self._x, self._y, self._z)

    def __repr__(self):
        return '<{:.6g}, {:.6g}, {:.6g}>'.format(self._x, self._y, self._z)

    def __add__(self, other):
        if type(other) is vector:
            return vector(self._x + other._x, self._y + other._y, self._z + other._z)
        return NotImplemented

    def __sub__(self, other):
        if type(other) is vector:
            return vector(self._x - other._x, self._y - other._y, self._z - other._z)
        return NotImplemented

    def __truediv__(self, other): # used by Python 3, and by Python 2 in the presence of __future__ division
        if isinstance(other, (float, int)):
            return vector(self._x / other, self._y / other, self._z / other)
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            return vector(self._x * other, self._y * other, self._z * other)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, (float, int)):
            return vector(self._x * other, self._y * other, self._z * other)
        return NotImplemented

    def __eq__(self,other):
        if type(self) is vector and type(other) is vector:
            return self.equals(other)
        return False

    def __ne__(self,other):
        if type(self) is vector and type(other) is vector:
            return not self.equals(other)
        return True

    @property
    def x(self):
        return self._x
    @x.setter
    def x(self,value):
        self._x = value
        self.on_change()

    @property
    def y(self):
        return self._y
    @y.setter
    def y(self,value):
        self._y = value
        self.on_change()

    @property
    def z(self):
        return self._z
    @z.setter
    def z(self,value):
        self._z = value
        self.on_change()

    @property
    def mag(self):
        return sqrt(self._x**2+self._y**2+self._z**2)
    @mag.setter
    def mag(self,value):
        normA = self.hat
        self._x = value * normA._x
        self._y = value * normA._y
        self._z = value * normA._z
        self.on_change()

    @property
    def mag2(self):
        return self._x**2+self._y**2+self._z**2
    @mag2.setter
    def mag2(self,value):
        normA = self.hat
        v = sqrt(value)
        self._x = v * normA._x
        self._y = v * normA._y
        self._z = v * normA._z
        self.on_change()

    @property
    def hat(self):
        smag = self.mag
        if ( smag > 0. ):
            return self / smag
        else:
            return vector(0., 0., 0.)
    @hat.setter
    def hat(self, value):
        smg = self.mag
        normA = value.hat
        self._x = smg * normA._x
        self._y = smg * normA._y
        self._z = smg * normA._z
        self.on_change()

    def norm(self):
        return self.hat

    def dot(self,other):
        return ( self._x*other._x + self._y*other._y + self._z*other._z )

    def cross(self,other):
        return vector( self._y*other._z-self._z*other._y,
                       self._z*other._x-self._x*other._z,
                       self._x*other._y-self._y*other._x )

    def proj(self,other):
        normB = other.hat
        return self.dot(normB) * normB

    def equals(self,other):
        return self._x == other._x and self._y == other._y and self._z == other._z

    def comp(self,other):  ## result is a scalar
        normB = other.hat
        return self.dot(normB)

    def diff_angle(self, other):
        a = self.hat.dot(other.hat)
        if a > 1:  # avoid roundoff problems
            return 0
        if a < -1:
            return pi
        return acos(a)

    def rotate(self, angle=0., axis=None):
        if axis is None:
            u = vector(0,0,1)
        else:
            u = axis.hat
        c = cos(angle)
        s = sin(angle)
        t = 1.0 - c
        x = u._x
        y = u._y
        z = u._z
        m11 = t*x*x+c
        m12 = t*x*y-z*s
        m13 = t*x*z+y*s
        m21 = t*x*y+z*s
        m22 = t*y*y+c
        m23 = t*y*z-x*s
        m31 = t*x*z-y*s
        m32 = t*y*z+x*s
        m33 = t*z*z+c
        sx = self._x
        sy = self._y
        sz = self._z
        return vector( (m11*sx + m12*sy + m13*sz),
                    (m21*sx + m22*sy + m23*sz),
                    (m31*sx + m32*sy + m33*sz) )

    def rotate_in_place(self, angle=0., axis=None):
        if axis is None:
            u = vector(0,0,1)
        else:
            u = axis.hat
        c = cos(angle)
        s = sin(angle)
        t = 1.0 - c
        x = u._x
        y = u._y
        z = u._z
        m11 = t*x*x+c
        m12 = t*x*y-z*s
        m13 = t*x*z+y*s
        m21 = t*x*y+z*s
        m22 = t*y*y+c
        m23 = t*y*z-x*s
        m31 = t*x*z-y*s
        m32 = t*y*z+x*s
        m33 = t*z*z+c
        sx = self._x
        sy = self._y
        sz = self._z
        self._x = m11*sx + m12*sy + m13*sz
        self._y = m21*sx + m22*sy + m23*sz
        self._z = m31*sx + m32*sy + m33*sz

def object_rotate(objaxis, objup, angle, axis):
    u = axis.hat
    c = cos(angle)
    s = sin(angle)
    t = 1.0 - c
    x = u._x
    y = u._y
    z = u._z
    m11 = t*x*x+c
    m12 = t*x*y-z*s
    m13 = t*x*z+y*s
    m21 = t*x*y+z*s
    m22 = t*y*y+c
    m23 = t*y*z-x*s
    m31 = t*x*z-y*s
    m32 = t*y*z+x*s
    m33 = t*z*z+c
    sx = objaxis._x
    sy = objaxis._y
    sz = objaxis._z
    objaxis._x = m11*sx + m12*sy + m13*sz # avoid creating a new vector object
    objaxis._y = m21*sx + m22*sy + m23*sz
    objaxis._z = m31*sx + m32*sy + m33*sz
    sx = objup._x
    sy = objup._y
    sz = objup._z
    objup._x = m11*sx + m12*sy + m13*sz
    objup._y = m21*sx + m22*sy + m23*sz
    objup._z = m31*sx + m32*sy + m33*sz

def mag(A):
    return A.mag

def mag2(A):
    return A.mag2

def norm(A):
    return A.hat

def hat(A):
    return A.hat

def dot(A,B):
    return A.dot(B)

def cross(A,B):
    return A.cross(B)

def proj(A,B):
    return A.proj(B)

def comp(A,B):
    return A.comp(B)

def diff_angle(A,B):
    return A.diff_angle(B)

def rotate(A, angle=0., axis = None):
    return A.rotate(angle,axis)

def adjust_up(oldaxis, newaxis, up, save_oldaxis): # adjust up when axis is changed
    if abs(newaxis._x) + abs(newaxis._y) + abs(newaxis._z) == 0:
        # If axis has changed to <0,0,0>, must save the old axis to restore later
        if save_oldaxis is None: save_oldaxis = oldaxis
        return save_oldaxis
    if save_oldaxis is not None:
        # Restore saved oldaxis now that newaxis is nonzero
        oldaxis._x = save_oldaxis._x # avoid creating a new vector
        oldaxis._y = save_oldaxis._y
        oldaxis._z = save_oldaxis._z
        save_oldaxis = None
    if newaxis.dot(up) != 0: # axis and up not orthogonal
        angle = oldaxis.diff_angle(newaxis)
        if angle > 1e-6: # smaller angles lead to catastrophes
            # If axis is flipped 180 degrees, cross(oldaxis,newaxis) is <0,0,0>:
            if abs(angle-pi) < 1e-6:
                up._x = -up._x
                up._y = -up._y
                up._z = -up._z
            else:
                rotaxis = oldaxis.cross(newaxis)
                up.rotate_in_place(angle=angle, axis=rotaxis) # avoid creating a new vector
    oldaxis._x = newaxis._x # avoid creating a new vector
    oldaxis._y = newaxis._y
    oldaxis._z = newaxis._z

def adjust_axis(oldup, newup, axis, save_oldup): # adjust axis when up is changed
    if abs(newup._x) + abs(newup._y) + abs(newup._z) == 0:
        # If up will be set to <0,0,0>, must save the old up to restore later
        if save_oldup is None: save_oldup = oldup
        return save_oldup
    if save_oldup is not None:
        # Restore saved oldup now that newup is nonzero
        oldup = save_oldup
        save_oldup = None
    if newup.dot(axis) != 0: # axis and up not orthogonal
        angle = oldup.diff_angle(newup)
        if angle > 1e-6: # smaller angles lead to catastrophes
            # If up is flipped 180 degrees, cross(oldup,newup) is <0,0,0>:
            if abs(angle-pi) < 1e-6:
                axis._x = -axis._x
                axis._y = -axis._y
                axis._z = -axis._z
            else:
                rotaxis = oldup.cross(newup)
                axis.rotate_in_place(angle=angle, axis=rotaxis) # avoid creating a new vector
    oldup._x = newup._x # avoid creating a new vector
    oldup._y = newup._y
    oldup._z = newup._z
    return save_oldup



from vpython import *

# Hard-sphere gas.

# Bruce Sherwood

win = 500

Natoms = 200  # change this to have more or fewer atoms

# Typical values
L = 1 # container is a cube L on a side
gray = color.gray(0.7) # color of edges of container
mass = 4E-3/6E23 # helium mass
Ratom = 0.03 # wildly exaggerated size of helium atom
k = 1.4E-23 # Boltzmann constant
T = 300 # around room temperature
dt = 1E-5

animation = canvas( width=win, height=win)
animation.range = L
animation.title = 'A "hard-sphere" gas'
s = """&nbsp;&nbsp;Theoretical and averaged speed distributions (meters/sec).
&nbsp;&nbsp;Initially all atoms have the same speed, but collisions
&nbsp;&nbsp;change the speeds of the colliding atoms. One of the atoms is
&nbsp;&nbsp;marked and leaves a trail so you can follow its path.
&nbsp
"""
animation.caption = s

d = L/2+Ratom
r = 0.005
boxbottom = curve(color=gray, radius=r)
boxbottom.append([vector(-d,-d,-d), vector(-d,-d,d), vector(d,-d,d), vector(d,-d,-d), vector(-d,-d,-d)])
boxtop = curve(color=gray, radius=r)
boxtop.append([vector(-d,d,-d), vector(-d,d,d), vector(d,d,d), vector(d,d,-d), vector(-d,d,-d)])
vert1 = curve(color=gray, radius=r)
vert2 = curve(color=gray, radius=r)
vert3 = curve(color=gray, radius=r)
vert4 = curve(color=gray, radius=r)
vert1.append([vector(-d,-d,-d), vector(-d,d,-d)])
vert2.append([vector(-d,-d,d), vector(-d,d,d)])
vert3.append([vector(d,-d,d), vector(d,d,d)])
vert4.append([vector(d,-d,-d), vector(d,d,-d)])

Atoms = []
p = []
apos = []
pavg = sqrt(2*mass*1.5*k*T) # average kinetic energy p**2/(2mass) = (3/2)kT

for i in range(Natoms):
    x = L*random()-L/2
    y = L*random()-L/2
    z = L*random()-L/2
    if i == 0:
        Atoms.append(simple_sphere(pos=vector(x,y,z), radius=Ratom, color=color.cyan, make_trail=True, retain=100, trail_radius=0.3*Ratom))
    else: Atoms.append(simple_sphere(pos=vector(x,y,z), radius=Ratom, color=gray))
    apos.append(vec(x,y,z))
    theta = pi*random()
    phi = 2*pi*random()
    px = pavg*sin(theta)*cos(phi)
    py = pavg*sin(theta)*sin(phi)
    pz = pavg*cos(theta)
    p.append(vector(px,py,pz))

deltav = 100 # binning for v histogram

def barx(v):
    return int(v/deltav) # index into bars array

nhisto = int(4500/deltav)
histo = []
for i in range(nhisto): histo.append(0.0)
histo[barx(pavg/mass)] = Natoms

gg = graph( width=win, height=0.4*win, xmax=3000, fast=False,
           xtitle='speed, m/s', ytitle='Number of atoms', ymax=Natoms*deltav/1000)

theory = gcurve( color=color.cyan, label='Theory' )
dv = 10
for v in range(0,3001+dv,dv):  # theoretical prediction
    theory.plot( v, (deltav/dv)*Natoms*4*pi*((mass/(2*pi*k*T))**1.5) *exp(-0.5*mass*(v**2)/(k*T))*(v**2)*dv )

accum = []
for i in range(int(3000/deltav)): accum.append([deltav*(i+.5),0])
vdist = gvbars(color=color.red, delta=deltav )

def interchange(v1, v2):  # remove from v1 bar, add to v2 bar
    barx1 = barx(v1)
    barx2 = barx(v2)
    if barx1 == barx2:  return
    histo[barx1] -= 1
    histo[barx2] += 1

def process_collisions(hitlist):
    # If any collisions took place, update momenta of the two atoms
    for ij in hitlist:
        i = ij[0]
        j = ij[1]
        ptot = p[i]+p[j]
        posi = apos[i]
        posj = apos[j]
        vi = p[i]/mass
        vj = p[j]/mass
        vrel = vj-vi
        a = vrel.mag2
        if a == 0: continue;  # exactly same velocities
        rrel = posi-posj
        b = 2*rrel.dot(vrel)
        c = rrel.mag2-4*Ratom*Ratom
        d = b*b-4*a*c
        if d < 0: continue  # something wrong; ignore this rare case
        deltat = (-b+sqrt(d))/(2*a) # t-deltat is when they made contact
        posi = posi-vi*deltat # back up to contact configuration
        posj = posj-vj*deltat
        mtot = 2*mass
        pcmi = p[i]-ptot*mass/mtot # transform momenta to cm frame
        pcmj = p[j]-ptot*mass/mtot
        rrel = norm(rrel)
        pcmi = pcmi-2*pcmi.dot(rrel)*rrel # bounce in cm frame
        pcmj = pcmj-2*pcmj.dot(rrel)*rrel
        p[i] = pcmi+ptot*mass/mtot # transform momenta back to lab frame
        p[j] = pcmj+ptot*mass/mtot
        apos[i] = posi+(p[i]/mass)*deltat # move forward deltat in time
        apos[j] = posj+(p[j]/mass)*deltat
        interchange(vi.mag, p[i].mag/mass)
        interchange(vj.mag, p[j].mag/mass)

def checkCollisions():
    hitlist = []
    r2 = 2*Ratom
    r2 *= r2
    for i in range(Natoms):
        ai = apos[i]
        for j in range(i+1, Natoms) :
            aj = apos[j]
            dr = ai - aj
            if mag2(dr) < r2: hitlist.append([i,j])
    if (len(hitlist) > 0): process_collisions(hitlist)

nhisto = 0 # number of histogram snapshots to average
timer = clock()

while True:
    rate(1000)
    # Accumulate and average histogram snapshots
    for i in range(len(accum)): accum[i][1] = (nhisto*accum[i][1] + histo[i])/(nhisto+1)
    if nhisto % 10 == 0:
        vdist.data = accum
    nhisto += 1

    # Update all positions
    for i in range(Natoms): Atoms[i].pos = apos[i] = apos[i] + (p[i]/mass)*dt

    # Check for collisions
    hitlist = checkCollisions()


    for i in range(Natoms):
        loc = apos[i]
        if abs(loc.x) > L/2:
            if loc.x < 0: p[i].x =  abs(p[i].x)
            else: p[i].x =  -abs(p[i].x)

        if abs(loc.y) > L/2:
            if loc.y < 0: p[i].y = abs(p[i].y)
            else: p[i].y =  -abs(p[i].y)

        if abs(loc.z) > L/2:
            if loc.z < 0: p[i].z =  abs(p[i].z)
            else: p[i].z =  -abs(p[i].z)

timer = clock()-timer
print(timer)



from vpython import *

# Gyroscope sitting on a pedestal

# The analysis is in terms of Lagrangian mechanics.
# The Lagrangian variables are polar angle theta,
# azimuthal angle phi, and spin angle psi.

scene.width = 800
scene.height = 600
scene.range = 1.2
scene.title = "A precessing, nutating gyroscope"
scene.visible = False

Lshaft = 1 # length of gyroscope shaft
r = Lshaft/2 # distance from support to center of mass
Rshaft = 0.03 # radius of gyroscope shaft
M = 1 # mass of gyroscope (massless shaft)
Rrotor = 0.4 # radius of gyroscope rotor
Drotor = 0.1 # thickness of gyroscope rotor
I3 = 0.5*M*Rrotor**2 # moment of inertia of gyroscope about its own axis
I1 = M*r**2 + .5*I3 # moment of inertia about a line through the support, perpendicular to the axis
hpedestal = Lshaft # height of pedestal
wpedestal = 0.1 # width of pedestal
tbase = 0.05 # thickness of base
wbase = 3*wpedestal # width of base
g = 9.8
Fgrav = vector(0,-M*g,0)
top = vector(0,0,0) # top of pedestal

shaft = cylinder(length=Lshaft, radius=Rshaft, color=color.orange)
rotor = cylinder(pos=vector(Lshaft/2-Drotor/2,0,0), axis=vector(Drotor, 0, 0),
                 radius=Rrotor, color=color.gray(0.9))
base = sphere(color=shaft.color, radius=Rshaft)
end = sphere(pos=vector(Lshaft,0,0), color=shaft.color, radius=Rshaft)
gyro = compound([shaft, rotor, base, end])
gyro_center = gyro.pos
gyro.texture = textures.metal
tip = sphere(pos=shaft.axis, radius=shaft.radius/2,  make_trail=True, retain=250)
tip.trail_color = color.green
tip.trail_radius = 0.15*Rshaft

pedestal = box(pos=top-vector(0,hpedestal/2+shaft.radius/2,0), 
               height=hpedestal-shaft.radius, length=wpedestal, width=wpedestal, texture=textures.wood)
pedestal_base = box(pos=top-vector(0,hpedestal+tbase/2,0), 
                    height=tbase, length=wbase, width=wbase, texture=textures.wood)

theta = 0
thetadot = 0
psi = 0
psidot = 0
phi = 0
phidot = 0
pureprecession = False

def reset():
    global theta, thetadot, psi, psidot, phi, phidot
    theta = 0.3*pi # initial polar angle of shaft (from vertical)
    thetadot = 0 # initial rate of change of polar angle
    psi = 0 # initial spin angle
    psidot = 30 # initial rate of change of spin angle (spin ang. velocity)
    phi = -pi/2 # initial azimuthal angle
    phidot = 0 # initial rate of change of azimuthal angle
    if pureprecession: # Set to True if you want pure precession, without nutation
        a = (1-I3/I1)*sin(theta)*cos(theta)
        b = -(I3/I1)*psidot*sin(theta)
        c = M*g*r*sin(theta)/I1
        phidot = (-b+sqrt(b**2-4*a*c))/(2*a)
    gyro.axis = gyro.length*vector(sin(theta)*sin(phi),cos(theta),sin(theta)*cos(phi))
    A = norm(gyro.axis)
    gyro.pos = 0.5*Lshaft*A
    tip.pos = Lshaft*A
    tip.clear_trail()

reset()
scene.waitfor('textures')
scene.visible = True

dt = 0.0001
t = 0
Nsteps = 20 # number of calculational steps between graphics updates

while True:
    rate(200)
    for step in range(Nsteps): # multiple calculation steps for accuracy
        # Calculate accelerations of the Lagrangian coordinates:
        atheta = sin(theta)*cos(theta)*phidot**2+( M*g*r*sin(theta)-I3*(psidot+phidot*cos(theta))*phidot*sin(theta))/I1
        aphi = (I3/I1)*(psidot+phidot*cos(theta))*thetadot/sin(theta)-2*cos(theta)*thetadot*phidot/sin(theta)
        apsi = phidot*thetadot*sin(theta)-aphi*cos(theta)
        # Update velocities of the Lagrangian coordinates:
        thetadot += atheta*dt
        phidot += aphi*dt
        psidot += apsi*dt
        # Update Lagrangian coordinates:
        theta += thetadot*dt
        phi += phidot*dt
        psi += psidot*dt

    gyro.axis = gyro.length*vector(sin(theta)*sin(phi),cos(theta),sin(theta)*cos(phi))
    # Display approximate rotation of rotor and shaft:
    gyro.rotate(angle=psidot*dt*Nsteps)
    A = norm(gyro.axis)
    gyro.pos = 0.5*Lshaft*A
    tip.pos = Lshaft*A
    t = t+dt*Nsteps
    



from vpython import *
from numpy import arange, clip

# David Scherer

scene.caption = """Right button drag or Ctrl-drag to rotate "camera" to view scene.
Middle button to drag or Alt-drag to zoom in or out.
  On a two-button mouse, middle is left + right."""
scene.title = "Lorenz differential equation"
scene.center = vector(25,0,0)
scene.background = color.white
scene.forward = vector(0,-.2,-1)
scene.range = 35

lorenz = curve( color = color.black, radius=0.3 )

# Draw grid
for x in arange(0,51,10):
    box(pos=vector(x,0,0), axis=vector(0,0,50), height=0.4, width=0.4, color=color.gray(0.6) )
for z in arange(-25,26,10):
    box(pos=vector(25,0,z), axis=vector(50,0,0), height=0.4, width=0., color=color.gray(0.6) )

dt = 0.01
y = vector(35, -10, -7)

for t in arange(0,10,dt):
    # Integrate a funny differential equation
    dydt = vector( - 8.0/3*y.x           +   y.y*y.z,
                              - 10*y.y +   10*y.z,
                 -  y.y*y.x +   28*y.y -   y.z );
    y = y + dydt*dt

    # Draw lines colored by speed
    c = clip( [mag(dydt) * 0.005], 0, 1 )[0]

    lorenz.append( pos=y, color=vector(c,0, 1-c) )

    rate( 500 )



from __future__ import division, print_function
"""
This module exports a VPython scene as POV-Ray scene description code.
Lights and camera location from the current scene are included.
Optionally, you may specify a list of include files,
   and pov textures for objects.
For an example of its use, see 'povexample.py'.
Objects not urrently exported:
    compound, 3D text, extrusion, points, triangle, quad
Further documentation is found at the start of the file.
"""

# This module exports a description of a VPython scene as a .pov file that can be
# read by the multiplatform free program POV-Ray, which produces a high-quality
# ray-traced image, with shadows (by default). The output is a png file (.png)
# which can be further processed by many programs, including Photoshop.

# To use, import the export routine from this file ("from povexport import export").
# When you have the VPython scene you want, execute "export()", then read the resulting
# .pov file into POV-ray.

# Here are the export routine's default options:
# export(canvas=None, filename=None, include_list=None, custom_text='', shadowless=0)
# If no canvas is specified, the current display is used.
# If no filename is specified, the title of your display will be used for the .pov file name.
# include_list lets you add your own POV-ray include statements to the .pov file.
# custom_text lets you add any kind of POV-Ray statements to the .pov file.
# shadowless=True makes POV-Ray not produce any shadows.
# You can give an individual object the attribute no_shadow,
#    which if True means it casts no shadow.
# You can give an individual object the attribute no_reflection,
#    which if True means it is not reflected by anything.

#--------------------------------------------------------------------------

# Ruth Chabay created the first version in 2000-12-17
# Contributors include Markus Gritsch, Bruce Sherwood, Scott David Daniels, Guy Kloss, and John Denker
# For details of their contributions, see the version of povexport for Classic VPython

# A version for VPython 7 created by Bruce Sherwood 2017-08-08

## NOTE: when changing this module please change the following string:
POVEXPORT_VERSION = "povexport 2017"

from vpython import *
from numpy import arange
import os

# Not handled: clone, compound, extrusion, 3D text

legal = {sphere:'sphere', box:'box', cylinder:'cylinder', helix:'helix',
         curve:'curve', ring:'ring', arrow:'arrow', label:'label',
         cone:'cone', ellipsoid:'ellipsoid', pyramid:'pyramid',
         distant_light:'distant_light', local_light:'local_light'}

ihat = vec(1, 0, 0)
jhat = vec(0, 1, 0)
khat = vec(0, 0, 1)
displayscale = 1.0 # global scale factor to adjust display.range to 100

def version():
    return POVEXPORT_VERSION

def getpolar(a):
    # a is a vec
    # find rotation angles (standard polar coord)
    xy = sqrt(a.x**2 + a.y**2)
    theta = atan2(xy, a.z)
    phi = atan2(a.y, a.x)
    return [theta, phi]

def find_rotations(a):
    # find rotations
    theta, phi = getpolar(a.axis)
    # find rotation around x-axis (if a.up <> jhat)
    # "undo" theta & phi rotations so can find alpha
    aup = vec(a.up)
    aup = aup.rotate(axis=khat, angle=-phi)
    aup = aup.rotate(axis=jhat, angle=pi/2-theta)
    a_sin = dot(cross(jhat, norm(aup)), ihat)
    a_cos = dot(norm(aup), jhat)
    alpha = atan2(a_sin, a_cos)
    return (alpha, theta, phi)

'''
def process_frame(a, code):
    # add in frame rotations & translations (may be nested)
    frame_code = ''
    fr = a.frame
    while fr:
        alpha, theta, phi = find_rotations(fr)
        frx=alpha*180./pi
        fry=-90+theta*180./pi
        frz=phi*180./pi
        rrot = '    rotate <%f, %f, %f>\n'
        ttrn = '    translate <%f, %f, %f>\n'
        frame_code += rrot % (frx, fry, frz)
        frame_code += ttrn % (displayscale*fr.x, displayscale*fr.y, displayscale*fr.z)
        fr = fr.frame

    # insert frame_code at end (these rot's must be done last)
    end = code.rfind('}')
    code = code[:end] + frame_code + code[end:]
    return code
'''

def add_texture(a, code):
    # add in user-specified POV-ray texture (replaces color)
    tex = None
    if hasattr(a, 'pov_texture'):
        tex = a.pov_texture
    if tex:
        start = code.rfind('pigment {')
        end = start + code[start:].find('}') + 1
        code = code[:start] + tex + code[end:] 
    return code

def no_shadow(a):
    if hasattr(a,"no_shadow") and a.no_shadow:
        return "no_shadow"
    else:
        return ""

def no_reflection(a):
    if hasattr(a,"no_reflection") and a.no_reflection:
        return "no_reflection"
    else:
        return ""

def transparency(a):
    if hasattr(a,"opacity"):
        return 1-a.opacity
    else:
        return 0

def export_sphere(a):
    sphere_template = """
sphere {
    <%(posx)f, %(posy)f, %(posz)f>, %(radius)f
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    %(no_shadow)s
    %(no_reflection)s
}
"""
    object_code = sphere_template % { 'posx':displayscale*a.pos.x, 'posy':displayscale*a.pos.y, 'posz':displayscale*a.pos.z,
                                      'radius':displayscale*a.radius, 'shininess':0.6, 
                                      'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                                      'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a)}
    object_code = add_texture(a, object_code)
    return object_code

def export_ellipsoid(a):
    ellipsoid_template = """
sphere {
    <0, 0, 0>, %(radius)f
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    scale <%(sizex)f, %(sizey)f, %(sizez)f>
    rotate <%(rotx)f, %(roty)f, %(rotz)f>
    translate <%(posx)f, %(posy)f, %(posz)f>
    %(no_shadow)s
    %(no_reflection)s
}
"""
    alpha, theta, phi = find_rotations(a)
    object_code = ellipsoid_template % { 'posx':displayscale*a.pos.x, 'posy':displayscale*a.pos.y, 'posz':displayscale*a.pos.z,
                    'radius':displayscale*a.size.x/2, 'shininess':0.6,
                    'sizex':1, 'sizey':a.size.y/a.size.x, 'sizez':a.size.z/a.size.x,
                    'rotx':alpha*180/pi, 'roty':-90+theta*180/pi, 'rotz':phi*180/pi,
                    'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                    'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a) }
    object_code = add_texture(a, object_code)
    return object_code

def export_box(a):
    # create box at origin along x-axis
    # then rotate around x,y,z axes
    # then translate to final location
    box_template = """
box {
    <%(posx)f, %(posy)f, %(posz)f>, <%(pos2x)f, %(pos2y)f, %(pos2z)f>
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    rotate <%(rotx)f, %(roty)f, %(rotz)f>
    translate <%(transx)f, %(transy)f, %(transz)f>
    %(no_shadow)s
    %(no_reflection)s
}
"""
    alpha, theta, phi = find_rotations(a)
    # pos of box is at center
    # generate two opposite corners for POV-ray
    pos1 = displayscale*(-a.size/2)
    pos2 = displayscale*( a.size/2)

    object_code = box_template % { 'posx':pos1.x, 'posy':pos1.y, 'posz':pos1.z,
                                   'pos2x':pos2.x, 'pos2y':pos2.y, 'pos2z':pos2.z,
                                   'rotx':alpha*180/pi, 'roty':-90+theta*180/pi, 'rotz':phi*180/pi,
                                   'transx':displayscale*a.pos.x, 'transy':displayscale*a.pos.y, 'transz':displayscale*a.pos.z,
                                   'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                                   'shininess':0.6,
                                   'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a) }
    object_code = add_texture(a, object_code)
    return object_code

def export_cylinder(a):
    cylinder_template = """
cylinder {
    <%(posx)f, %(posy)f, %(posz)f>,<%(pos2x)f, %(pos2y)f, %(pos2z)f>, %(radius)f
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    %(no_shadow)s
    %(no_reflection)s
}
"""
    endpt1=displayscale*a.pos
    endpt2=displayscale*(a.pos+a.axis)
    object_code = cylinder_template % { 'posx':endpt1.x, 'posy':endpt1.y, 'posz':endpt1.z,
                                        'pos2x':endpt2.x, 'pos2y':endpt2.y, 'pos2z':endpt2.z,
                                        'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                                        'radius':displayscale*a.radius, 'shininess':0.6,
                                        'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a) }
    object_code = add_texture(a, object_code)
    return object_code

def export_helix(a):
    thickness = a.thickness
    pts = []
    N = 40 # number of curve segments in one coil
    L = a.length
    R = a.radius
    dx = L/(N*a.coils)
    for x in arange(0,L+dx/2,dx):
        pts.append(vec(x, R*sin(2*pi*(x/(N*dx))), R*cos(2*pi*(x/(N*dx)))))
    # Create a curve whose size is LxRxR:
    c = curve(pos=pts, color=a.color, radius=thickness/2, visible=False)
    c.origin = a.pos
    c.axis = a.axis
    c.up = a.up
    return export_curve(c)

def export_curve(a):
    object_code = ''
    ccyl = cylinder(visible=False) # create cylinder and sphere that can be deleted
    csph = sphere(visible=False)
    curve_no_shadow = no_shadow(a)
    curve_no_reflection = no_reflection(a)
    rotaxis = vec(1,0,0).cross(a.axis)
    rotangle = diff_angle(vec(1,0,0),a.axis)
    if a.npoints > 1:
        ii = 0
        while ii < a.npoints-1:
            endpt1 = a.point(ii)
            endpt2 = a.point(ii+1)
            if endpt1['radius'] == 0:
                rr = a.radius
            else:
                rr = endpt1['radius']
            curve_color = endpt1['color']
            if ii == 0:
                csph = sphere(pos=a.origin+endpt1['pos'], radius=rr, color=curve_color,
                          no_shadow=curve_no_shadow,
                          no_reflection=curve_no_reflection, visible=False)
                object_code += export_sphere(csph)
            if endpt2 == endpt1:
                ii += 1
                continue
            p1 = vec(endpt1['pos']) # make copies of these position vectors
            p2 = vec(endpt2['pos'])
            if rotangle > 0.001:
                if abs(rotangle-pi) < 0.001:
                    p1 = - p1
                    p2 = - p2
                else:
                    p1 = p1.rotate(angle=rotangle, axis=rotaxis)
                    p2 = p2.rotate(angle=rotangle, axis=rotaxis)
            ccyl = cylinder(pos=a.origin+p1, axis=(p2-p1),
                            radius=rr, color=curve_color,
                            no_shadow=curve_no_shadow,
                            no_reflection=curve_no_reflection, visible=False)
            csph = sphere(pos=a.origin+p2, radius=rr, color=curve_color,
                          no_shadow=curve_no_shadow,
                          no_reflection=curve_no_reflection, visible=False)
            if hasattr(a, 'pov_texture'):
                ccyl.pov_texture = a.pov_texture
                csph.pov_texture = a.pov_texture
            object_code += export_cylinder(ccyl) + export_sphere(csph)
            ii += 1
        if hasattr(a, 'pov_texture'):
            csph.pov_texture = a.pov_texture
    return object_code

def export_ring(a):
    torus_template = """
torus {
    %(radius)f, %(minorradius)f
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    rotate <0.0, 0.0, -90.0> // align with x-axis
    rotate <%(rotx)f, %(roty)f, %(rotz)f>
    translate <%(transx)f, %(transy)f, %(transz)f>
    %(no_shadow)s
    %(no_reflection)s
}
"""
    ang = getpolar(a.axis)
    theta = ang[0]
    phi = ang[1]
    object_code = torus_template % { 'radius':displayscale*a.radius, 'minorradius':displayscale*a.thickness,
                                     'transx':displayscale*a.pos.x, 'transy':displayscale*a.pos.y, 'transz':displayscale*a.pos.z,
                                     'rotx':0, 'roty':-90+theta*180/pi, 'rotz':phi*180/pi, 'shininess':0.6,
                                     'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                                     'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a) }
    object_code = add_texture(a, object_code)
    return object_code

def export_pyramid(a):
    pyramid_template = """
object {Pyramid
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    scale <%(scalex)f, %(scaley)f, %(scalez)f>
    rotate <%(rotx)f, %(roty)f, %(rotz)f>
    translate <%(transx)f, %(transy)f, %(transz)f>
    %(no_shadow)s
    %(no_reflection)s
}
"""
    alpha, theta, phi = find_rotations(a)
    object_code = pyramid_template % { 'scalex':displayscale*a.size.x,
                                       'scaley':displayscale*a.size.y,
                                       'scalez':displayscale*a.size.z,
                              'rotx':0, 'roty':-90+theta*180/pi, 'rotz':phi*180/pi, 'shininess':0.6,
                              'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                              'transx':displayscale*a.pos.x, 'transy':displayscale*a.pos.y, 'transz':displayscale*a.pos.z,
                              'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a) }
    object_code = add_texture(a, object_code)
    return object_code

def export_arrow(a):
    al = a.length
    if al == 0:       ## omit zero-length arrows
        return
    sw = a.shaftwidth
    if sw == 0:
        sw = 0.1*al
    hw = a.headwidth
    if hw == 0:
        hw = 2*sw
    hl = a.headlength
    if hl == 0:
        hl = 3*sw
    if a.shaftwidth != 0:
        if hl > .5*al:
            hl = .5*al
    else:
        if sw < .02*al:
            swtemp = .02*al
            hw *= swtemp/sw
            hl *= swtemp/sw
            sw = swtemp
        if hl > .5*al:
            hltemp = .5*al
            hw *= hltemp/hl
            sw *= hltemp/hl
            hl = hltemp
    sl = al-hl # length of shaft
    arrow_no_shadow = no_shadow(a)
    arrow_no_reflection = no_reflection(a)

    # head is a pyramid; need to create a dummy pyramid
    apyramid = pyramid(pos=a.pos+a.axis.norm()*sl, up=a.up,
               size=vec(hl,hw,hw), color=a.color, no_shadow=arrow_no_shadow,
               shininess=a.shininess,
               no_reflection=arrow_no_reflection, opacity=a.opacity, visible=False)
    apyramid.axis = hl*a.axis.norm() 
    m1 = export_pyramid(apyramid)
    m1 = add_texture(a, m1)

    # shaft is a box; need to create a dummy box
    abox = box(pos=(a.pos+a.axis*(sl/al)/2), axis=(a.axis*(sl/al)),
               up=a.up, width=sw, height=sw,
               color=a.color, opacity=a.opacity, shininess=a.shininess,
               no_shadow=arrow_no_shadow, no_reflection=arrow_no_reflection, visible=False)
    m2 = export_box(abox)
    m2 = add_texture(a, m2)
    # concatenate pyramid & box
    object_code = m1 + m2
    return object_code

def export_cone(a):
    cone_template = """
cone {
    <%(posx)f, %(posy)f, %(posz)f>, %(radius)f
    <%(pos2x)f, %(pos2y)f, %(pos2z)f>, %(minorradius)f
    texture {
        pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
        finish { phong %(shininess)f }
    }
    %(no_shadow)s
    %(no_reflection)s
}
"""
    pos2 = displayscale*(a.pos+a.axis)
    object_code = cone_template % { 'radius':displayscale*a.radius, 'minorradius':0,
                                    'posx':displayscale*a.pos.x, 'posy':displayscale*a.pos.y, 'posz':displayscale*a.pos.z,
                                    'pos2x':pos2.x, 'pos2y':pos2.y, 'pos2z':pos2.z, 'shininess':0.6,
                                    'red':a.red, 'green':a.green, 'blue':a.blue, 'transparency':transparency(a),
                                    'no_shadow':no_shadow(a), 'no_reflection':no_reflection(a) }
    object_code = add_texture(a, object_code)
    return object_code

def export_label(a):
    label_template = """
text {
    ttf "cyrvetic.ttf" "%(text)s" 0.1, 0
    pigment {color rgbt <%(red)f, %(green)f, %(blue)f, %(transparency)f>}
    translate <%(transx)f, %(transy)f, %(transz)f>
    %(no_shadow)s
    %(no_reflection)s
}
"""
    object_code = label_template % {'transx': displayscale * a.pos.x,
                                    'transy': displayscale * a.pos.y,
                                    'transz': displayscale * a.pos.z,
                                    'red': a.color.x, 'green': a.color.y, 'blue': a.color.z, 'transparency': transparency(a),
                                    'text': a.text,
                                    'no_shadow': no_shadow(a), 'no_reflection': no_reflection(a)}
    object_code = add_texture(a, object_code)
    return object_code

def export(canvas=None, filename=None, include_list=None, custom_text='', shadowless=0):
    canv = canvas

    global displayscale
    if canv == None:         # no display specified so find active display
        b = box(visible=0)
        canv = b.canvas

    if filename == None:
        if len(canv.title) == 0: filename = 'povray.pov'
        else: filename = canv.title + '.pov'

    if include_list == None:
        include_text = ''
        # Maybe should always include the following definitions?
        #include_text = '#include "colors.inc"\n#include "stones.inc"\n#include "woods.inc"\n#include "metals.inc"\n'
    else:
        include_text = '\n'
        for x in include_list:
            include_text = include_text + '#include "'+ x + '"\n'

    initial_comment = """// POV-ray code generated by povexport.py
"""

    pyramid_def = """
// Four-sided pyramid from shapes2.inc, slightly modified.
// The 1x1 base of the pyramid is at the origin, in the yz plane.
// The tip of the pyramid is at <1,0,0>.
#declare Pyramid = union {
    triangle { <0, -0.5, -0.5>, <0, +0.5, -0.5>, <1, 0, 0> }
    triangle { <0, +0.5, -0.5>, <0, +0.5, +0.5>, <1, 0, 0> }
    triangle { <0, -0.5, +0.5>, <0, +0.5, +0.5>, <1, 0, 0> }
    triangle { <0, -0.5, +0.5>, <0, -0.5, -0.5>, <1, 0, 0> }
    triangle { <0, -0.5, -0.5>, <0, -0.5, +0.5>, <0, 0.5, 0.5> }
    triangle { <0, -0.5, -0.5>, <0, +0.5, -0.5>, <0, 0.5, 0.5> }
}
"""

    ambient_template = """
global_settings { ambient_light rgb <%(red)f, %(green)f, %(blue)f> }
"""

    scalar_ambient_template = """
global_settings { ambient_light rgb <%(amb)f, %(amb)f, %(amb)f> }
"""

    background_template = """
background { color rgb <%(red)f, %(green)f, %(blue)f> }
"""

    light_template = """
light_source { <%(posx)f, %(posy)f, %(posz)f>
    color rgb <%(red)f, %(green)f, %(blue)f>
}
"""

    camera_template = """
camera {
    right <-image_width/image_height, 0, 0>      // vpython uses right-handed coord. system
    location <%(posx)f, %(posy)f, %(posz)f>
    up <%(upx)f, %(upy)f, %(upz)f>
    look_at <%(pos2x)f, %(pos2y)f, %(pos2z)f>
    angle %(fov)f
}
"""

    # begin povray setup
    file = open(filename, 'w')
    print(os.path.realpath(file.name))

    file.write( initial_comment + include_text + custom_text + pyramid_def )
    file.write( ambient_template % { 'red':canv.ambient.x*10 ,
                                         'green':canv.ambient.y*10,
                                         'blue':canv.ambient.z*10 })
    file.write( background_template % { 'red':canv.background.x,
                                        'green':canv.background.y,
                                        'blue':canv.background.z } )

    displayscale = 10/canv.range # deal with very small range values (e.g. atomic sizes)

    for light in canv.lights: # reproduce vpython lighting (not ideal, but good approximation)
        if type(light) is distant_light:
            pos = norm(light.direction) * 1000 # far away to simulate parallel light
        elif type(light) is local_light:
            pos = displayscale*light.pos
        lcolor = light.color
        light_code = light_template % { 'posx':pos.x, 'posy':pos.y, 'posz':pos.z,
                                        'red':lcolor.x, 'green':lcolor.y, 'blue':lcolor.z }
        if shadowless:
            end = light_code.rfind('}')
            light_code = light_code[:end] + '    shadowless\n' + light_code[end:]
        file.write( light_code )

    cpos = 1.5*displayscale*canv.camera.pos # 1.5 is a not understood fudge factor
    ctr = displayscale*canv.center
    cup = canv.up
    file.write( camera_template % { 'posx':cpos.x, 'posy':cpos.y, 'posz':cpos.z,
                                    'upx':cup.x, 'upy':cup.y, 'upz':cup.z,
                                    'pos2x':ctr.x, 'pos2y':ctr.y, 'pos2z':ctr.z,
                                    'fov':canv.fov*180/pi } )

    for obj in canv.objects:
        key = obj.__class__
        if key in legal:
            obj_name = legal[key]
            if obj_name == 'distant_light' or obj_name == 'local_light': continue
            function_name = 'export_' + obj_name
            function = globals().get(function_name)
            object_code = function(obj)
            if object_code is None:
                continue
            file.write( object_code )
        else:
            print('WARNING: export function for ' + str(obj.__class__) + ' not implemented')

    file.close()
    return 'The export() function no longer returns the display as an\n' \
           'endless POV-Ray string, but saves it to a file instead.'
# end defining export()


if __name__ == '__main__':
    print(__doc__)



import os

import yaml
from binstar_client import Conflict

from sync import sync

FROM = 'conda-forge'
TO = 'vpython'


def main(requirements_path):
    with open(requirements_path, 'rt') as f:
        package_list = yaml.safe_load(f)

    token = os.getenv('BINSTAR_TOKEN')

    for package in package_list:
        print(package['name'])
        try:
            sync(FROM, TO, package['name'], token, to_label='main')
        except Conflict:
            print('Skipping because of a conflict anaconda.org')


if __name__ == '__main__':
    main('vp_copy.yaml')



import os
import re
from io import open


def glowscript_version():
    """
    Extract the Glowscript version from the javascript in the data directory.
    """

    this_dir = os.path.dirname(os.path.abspath(__file__))
    glowscript_lib = os.path.join(this_dir, 'vpython_libraries', 'glow.min.js')

    with open(glowscript_lib, 'r', encoding='UTF-8') as f:
        glow_contents = f.read()

    matches = re.search(r'glowscript={version:"(.*?)"}', glow_contents)

    if matches:
        gs_version = matches.group(1)
    else:
        raise RuntimeError("Could not determine glowscript version.")

    return gs_version



from vpython import *
scene.width = scene.height = 600
scene.range = 0.6

# A pulse ripples along a rug, demonstrating dynamic changea of shape
# Bruce Sherwood, May 2012

def display_instructions():
    s = "In GlowScript programs:\n\n"
    s += "    Rotate the camera by dragging with the right mouse button,\n        or hold down the Ctrl key and drag.\n\n"
    s += "    To zoom, drag with the left+right mouse buttons,\n         or hold down the Alt/Option key and drag,\n         or use the mouse wheel.\n"
    s += "\nTouch screen: pinch/extend to zoom, swipe or two-finger rotate."
    scene.caption = s
    
display_instructions()

# Construct a square WxH divided into little squares
# There are (w+1)x(h+1) vertices
# Center of rug is at 0,0,0

H = W = 1
w = 1
h = 50
dx = W/w
dy = H/h

# Create a grid of vertex objects covering the rug
verts = []
for y in range(h+1): # from 0 to h inclusive, to include both bottom and top edges
    verts.append([])
    for x in range(w+1): # from 0 to w inclusive, to include both left and right edges
        verts[y].append(vertex(pos=vector(-0.5+x*dx,-0.5+y*dy,0), normal=vector(0,0,1), texpos=vector(x/w,y/h,0), shininess= 0))

# Create quads (equivalent to two triangles) based on the vertex objects just created.
# Note that a particular vertex may be shared by as many as 4 neighboring quads, and
# changing one vertex affects all of the quads that use that vertex.
for y in range(h): # from 0 to h, not including h
    for x in range(w): # from 0 to w, not including w
        quad(vs=[verts[y][x], verts[y][x+1], verts[y+1][x+1], verts[y+1][x]], texture=textures.rug)

scene.waitfor('textures') # wait until the rug texture has been loaded

Lpulse = 0.4 # length of half sine wave
dy_pulse = Lpulse/50
k = pi/(0.6*Lpulse)
A = 0.05

def pulse(z): # return the pulse height and normal
    if z < 0.2*Lpulse: return 0
    if z > 0.8*Lpulse: return 0
    z -= 0.2*Lpulse
    return A*sin(k*z)

run = True

def down(ev):
    global run
    run = not run

scene.bind("mousedown", down)

y = -0.5-Lpulse-dy_pulse # bottom of pulse (starts below rug)
while True:
    rate(50)
    while not run:
        scene.waitfor('redraw')
    y += dy_pulse
    if y+Lpulse <= -0.5:
        continue
    if y >= 0.5:
        y = -0.5-Lpulse
        continue
        
    # Note: floor and ceil are floats in Python 2 but ints in Python 3
    start = int(floor((y+0.5)/dy))     # lowest row of vertices in pulse
    end = int(ceil((y+0.5+Lpulse)/dy)) # highest row of vertices in pulse
    if start < 0:
        if end <= 0: continue
        start = 0
    if end > h:
        end = h
    
    yp = -0.5+start*dy
    for s in range(start,end):
        z0 = pulse(yp-y-dy_pulse)
        z1 = pulse(yp-y)
        z2 = pulse(yp+dy_pulse-y)
        yp += dy # advance to next row
        
        # If slope of a line is dy/dz, normal to the line is in direction < -dz, +dy >
        n1y0 = -(z1-z0)
        n2y0 = -(z2-z1)
        n1y = .5*(n1y0+n2y0) # average adjacent normals to smooth the lighting
        n1z = dy
        
        vy = verts[s]
        for vx in range(w+1):
            vy[vx].pos.z = z1
            vy[vx].normal = vector(0,n1y,n1z) 



from vpython import *
b = box(pos=vec(-4,2,0), color=color.red)
c1 = cylinder(pos=b.pos, radius=0.1, axis=vec(0,1.5,0), color=color.yellow)
s = sphere(pos=vec(4,-4,0), radius=0.5, color=color.green)
c2 = cylinder(pos=s.pos, radius=0.1, axis=vec(0,1.5,0), color=color.yellow)
t1 = text(text='box', pos=c1.pos+c1.axis, align='center', height=0.5,
          color=color.yellow, billboard=True, emissive=True)
t2 = text(text='sphere', pos=c2.pos+c2.axis, align='center', height=0.5,
          color=color.yellow, billboard=True, emissive=True)
t3 = text(text='Faces forward', pos=vec(-4,0,0),
          color=color.cyan, billboard=True, emissive=True)
box(pos=t3.start, size=0.1*vec(1,1,1), color=color.red)
t4 = text(text='Regular text', pos=vec(-4,-1,0), depth=0.5, color=color.yellow,
        start_face_color=color.red, end_face_color=color.green)
box(pos=t4.start, size=0.1*vec(1,1,1), color=color.red)

scene.caption = """<b>3D text can be "billboard" text -- always facing you.</b>
Note that the "Regular text" has different colors on the front, back and sides.
Right button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
  On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""



from vpython import *

# Stars interacting gravitationally

# Bruce Sherwood

scene.width = scene.height = 600

# Display text below the 3D graphics:
scene.title = "Stars interacting gravitationally"

scene.caption = """Right button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
  On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""

Nstars = 20  # change this to have more or fewer stars

G = 6.7e-11 # Universal gravitational constant

# Typical values
Msun = 2E30
Rsun = 2E9
L = 4e10
vsun = 0.8*sqrt(G*Msun/Rsun)

scene.range = 2*L
scene.forward = vec(-1,-1,-1)

xaxis = curve(color=color.gray(0.5), radius=3e8)
xaxis.append(vec(0,0,0))
xaxis.append(vec(L,0,0))
yaxis = curve(color=color.gray(0.5), radius=3e8)
yaxis.append(vec(0,0,0))
yaxis.append(vec(0,L,0))
zaxis = curve(color=color.gray(0.5), radius=3e8)
zaxis.append(vec(0,0,0))
zaxis.append(vec(0,0,L))

Stars = []
star_colors = [color.red, color.green, color.blue,
              color.yellow, color.cyan, color.magenta]

psum = vec(0,0,0)
for i in range(Nstars):
    star = sphere(pos=L*vec.random(), make_trail=True, retain=150, trail_radius=3e8)
    R = Rsun/2+Rsun*random()
    star.radius = R
    star.mass = Msun*(R/Rsun)**3
    star.momentum = vec.random()*vsun*star.mass
    star.color = star.trail_color = star_colors[i % 6]
    Stars.append( star )
    psum = psum + star.momentum

#make total initial momentum equal zero
for i in range(Nstars):
    Stars[i].momentum = Stars[i].momentum - psum/Nstars

dt = 1000
hitlist = []

def computeForces():
    global hitlist, Stars
    hitlist = []
    N = len(Stars)
    for i in range(N):
        si = Stars[i]
        if si is None: continue
        F = vec(0,0,0)
        pos1 = si.pos
        m1 = si.mass
        radius = si.radius
        for j in range(N):
            if i == j: continue
            sj = Stars[j]
            if sj is None: continue
            r = sj.pos - pos1
            rmag2 = mag2(r)
            if rmag2 <= (radius+sj.radius)**2: hitlist.append([i,j])
            F = F + (G*m1*sj.mass/(rmag2**1.5))*r
        si.momentum = si.momentum + F*dt

while True:
    rate(100)
    
    # Compute all forces on all stars
    computeForces()

    # Having updated all momenta, now update all positions
    for star in Stars:
        if star is None: continue
        star.pos = star.pos + star.momentum*(dt/star.mass)

    # If any collisions took place, merge those stars
    hit = len(hitlist)-1
    while hit > 0:
        s1 = Stars[hitlist[hit][0]]
        s2 = Stars[hitlist[hit][1]]
        if (s1 is None) or (s2 is None): continue
    
        mass = s1.mass + s2.mass
        momentum = s1.momentum + s2.momentum
        pos = (s1.mass*s1.pos + s2.mass*s2.pos) / mass
        s1.color = s1.trail_color = (s1.mass*s1.color + s2.mass*s2.color) / mass
        R = Rsun*(mass / Msun)**(1/3)
        
        s1.clear_trail()
        s2.clear_trail()
        s2.visible = False
        
        s1.mass = mass
        s1.momentum = momentum
        s1.pos = pos
        s1.radius = R
        Stars[hitlist[hit][1]] = None
        hit -= 1



from vpython import *
from random import random
scene.width = scene.height = 600
scene.background = color.white
scene.range = 1.3
scene.title = 'A 15000-segment curve object with random segment lengths, radii, and colors'

run = True
def Runbutton(b):
    global run
    if b.text == 'Pause':
        run = False
        b.text = 'Run'
    else:
        run = True
        b.text = 'Pause'
        
button(text='Pause', bind=Runbutton)
scene.append_to_caption("""<br>Right button drag or Ctrl-drag to rotate "camera" to view scene.
Middle button or Alt-drag to drag up or down to zoom in or out.
  On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate.""")

N = 15000
p = []
last = vec(0,0,0)
for i in range(N):
    next = last+0.1*vec.random()
    while mag(next) > 1: # if next is outside the sphere, try another random value
        next = last+0.1*vec.random()
    p.append({'pos':next, 'radius':0.001+0.03*random(), 'color':(vec(1,1,1)+vec.random())/2})
    last = next
c = curve(p)
while True:
    rate(60)
    if run:
        c.rotate(angle=0.005, axis=vec(0,1,0))



from vpython import *
s = '    Test of <b><i>graphing</i></b>. Move the mouse over the graph to display data.'
test = graph(title=s, xtitle='time', ytitle='value', fast=False)
funct1 = gcurve(color=color.blue, width=4, dot=True, 
        dot_color=color.green, dot_radius=15, radius=5, markers=True, marker_color=color.cyan)
funct2 = gvbars(delta=0.4, color=color.red, label='bars')
funct3 = gdots(color=color.orange)

for t in range(-30, 74):
    rate(500)
    funct1.plot( t, 5.0+5.0*cos(-0.2*t)*exp(0.015*t) )
    funct2.plot( t, 2.0+5.0*cos(-0.1*t)*exp(0.015*t) )
    funct3.plot( t, 5.0*cos(-0.03*t)*exp(0.015*t) )

scene.width = scene.height = 100
box()
scene.pause()
test.title = 'New title'
test.xtitle = "New xtitle"
test.ytitle = 'New ytitle'
scene.pause()
funct2.data = [(-20,8), (-30,4)]
funct2.color = color.green
scene.pause()
funct2.label = 'new bars'
funct3.label = 'dots'
scene.pause()
funct2.legend = False
scene.pause()
funct2.delta = 10
funct1.width = 1
funct3.radius = 2*funct3.radius
scene.pause()
#funct3.size = 20
#funct2.horizontal = True
#funct2.orientation = 'h'
funct1.dot_radius = 40
funct1.dot_color = color.cyan
funct1.marker_color = color.black
scene.pause()
funct1.markers = False
scene.pause()
funct1.delete()
scene.pause()
test.delete()
print('done')


from vpython import *
scene.width = 600
scene.height = 600
show = 'box'
last_show = show
 
D = 0.7 # size of box
R = .4 # radius of sphere

names = ['flower', 'granite', 'gravel', 'metal', 'rock', 'rough', 'rug', 'stones', 'stucco', 'wood', 'wood_old', 'earth']
Ts = [textures.flower, textures.granite, textures.gravel, textures.metal, textures.rock, textures.rough,
      textures.rug, textures.stones, textures.stucco, textures.wood, textures.wood_old, textures.earth]
bumps = [ None, None, 'gravel', None, 'rock', None, None, 'stones', 'stucco', None, 'wood_old']

def erase():
    for obj in scene.objects:
        obj.visible = False

def show_object(index, x, y):
    T = Ts[index]
    B = None
    # Bump maps aren't very salient unless one moves the light or rotates the object,
    # so don't bother with bump maps unless there's an option to move the light or object.
    #if (bumps[index] !== None) B = bumpmaps[bumps[index]]
    if show == 'box':
        c = box(pos=vec(x,y,0), size=D*vec(1,1,1))
    elif show == 'sphere':
        c = sphere(pos=vec(x,y,0), size=D*vec(1,1,1))
    elif show == 'cylinder': 
        c = cylinder(pos=vec(x-D/2,y,0), size=D*vec(1,1,1))
    elif show == 'cone': 
        c = cone(pos=vec(x-D/2,y,0), size=D*vec(1,1,1))
    elif show == 'pyramid': 
        c = pyramid(pos=vec(x-D/2,y,0), size=D*vec(1,1,1))
    c.index = index
    c.shininess = 0
    c.texture = {'file':T, 'bumpmap':B}
    label(pos=vec(x,y-.5,0), box=0, text='textures.'+names[index])

def setup():
    scene.range = 2.2
    scene.fov = 0.2
    scene.center = vec(1.5,2,0)
    scene.forward = vec(0,0,-1)
    erase()
    scene.visible = False
    index = 0
    y = 3.3
    while y > 0:
        for x in range(4):
            if index >= len(names): break; 
            show_object(index, x, y)
            index += 1
        y -= 1.3
    scene.visible = True

setup()
scene.visible = False
scene.caption = "Loading textures..."
scene.waitfor("textures")
scene.caption = "Choose the type of object:  "
scene.visible = True

def choose(c):
    global show
    show = c.selected
    
menu(choices=['box', 'sphere', 'cylinder', 'cone', 'pyramid'], selected='box', bind=choose)

scene.append_to_caption('\n\nClick an object to enlarge it; then click anywhere to show all objects again.')

hit = None
clicked = False
def handle_click(ev):
    global hit, clicked
    hit = scene.mouse.pick
    clicked = True
scene.bind('click', handle_click)

def single_object(index):
    scene.center = vec(0,-.1*R,0)
    scene.range = 1.5*R
    erase()
    show_object(index, 0, 0)

picked = None
    
while True:
    rate(30)
    if show != last_show:
        last_show = show
        if picked != None:
            single_object(picked.index)
        else:
            setup()
    if clicked:
        clicked = False
        if picked != None:
            picked = None
            setup()
        elif picked == None and hit != None:
            picked = hit
            hit = None
            single_object(picked.index)



from vpython import *
scene.background = color.gray(0.8)
scene.forward = vec(0,-0.2,-1)
scene.fov = 0.2
scene.range = 3.8
scene.caption = """Right button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
     On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate.\n"""

E1 = extrusion(path=[vec(0,0,0), vec(0,0,-0.7)], texture=textures.wood_old,
    shape=[ shapes.circle(radius=1.5),
            shapes.triangle(pos=[0,-0.6], length=1.2),
            shapes.trapezoid(pos=[0,0.6], width=1.6,
              height=1, top=0.6) ], pos=vec(-4,1.5,0))

copper = vec(0.722,0.451,0.200)
E2 = extrusion(path=paths.arc(radius=1.7, angle2=pi), texture=textures.metal, 
    shape=[ [shapes.triangle(length=2), shapes.circle(pos=[0,.5], radius=0.2),
    shapes.trapezoid(pos=[0,-0.2], width=0.6, height=0.4)],
    [shapes.rectangle(pos=[0,1.8], width=1,height=0.3)] ],
    start_face_color=copper, end_face_color=copper)
E2center = E2.pos    # initial pos of center of extrusion
E2.pos = vec(3,2,0)  # new pos of center of extrsion
E2rot = E2.pos+vec(0,0,-E2center.z) # a location at the front of the extrusion

halo = ring(pos=vec(0,0,1), radius=0.8, thickness=0.2, color=color.cyan )

rect = extrusion(path=paths.rectangle(width=5, height=2), 
            shape=shapes.hexagon(length=0.3), color=color.red)

bottom = extrusion(path=paths.cross(width=4, thickness=1),
            shape=shapes.circle(radius=0.2), color=color.green)

tube = extrusion(path=[vec(0,0,0), vec(2,0,0)], shape=shapes.circle(radius=0.6, thickness=0.2),
                    pos=vec(-1,1.7,0), axis=vec(0,0,2), color=color.yellow, end_face_color=color.blue)
text(pos=tube.pos+vec(0,0,1), text='tube', align='center',
        height=0.25, depth=0, color=color.blue)

run = True

def runner(r):
    global run
    run = r.checked
    
checkbox(bind=runner, text='Run', checked=True)

scene.waitfor('textures')

t = 0
dt = 0.01
dtheta = 0.01
while True:
    rate(100)
    if run:
        halo.pos.x = -1+sin(2*t)
        bottom.pos.y = -2+0.5*cos(3*t)
        E1.rotate(angle=dtheta, axis=vec(0,1,0))
        E2.rotate(angle=-dtheta, axis=vec(0,1,0), origin=E2rot)
        t += dt




from vpython import *
scene.width = scene.height = 500
scene.background = color.gray(0.8)
scene.range = 2.2
scene.caption = "Click to pick an object and make it red."
scene.append_to_caption("\nNote picking of individual curve segments.")
box(pos=vector(-1,0,0), color=color.cyan, opacity=1)
box(pos=vector(1,-1,0), color=color.green)
arrow(pos=vector(-1,-1.3,0), color=color.orange)
cone(pos=vector(2,0,0), axis=vector(0,1,-.3), color=color.blue, size=vector(2,1,1))
sphere(pos=vector(-1.5,1.5,0), color=color.white, size=.4*vector(3,2,1))
square = curve(color=color.yellow, radius=.05)
square.append(vector(0,0,0))
square.append(pos=vector(0,1,0), color=color.cyan, radius=.1)
square.append(vector(1,1,0))
square.append(pos=vector(1,0,0), radius=.1)
square.append(vector(0.3,-.3,0))
v0 = vertex(pos=vector(-.5,1.2,0), color=color.green)
v1 = vertex(pos=vector(1,1.2,0), color=color.red)
v2 = vertex(pos=vector(1,2,0), color=color.blue)
v3 = vertex(pos=vector(-.5,2,0), color=color.yellow)
quad(vs=[v0, v1, v2, v3])
extrusion(path=[vector(-1.8,-1.3,0), vector(-1.4,-1.3,0)],
            shape=shapes.circle(radius=.5, thickness=0.4), color=color.yellow)
ring(pos=vector(-0.6,-1.3,0), size=vector(0.2,1,1), color=color.green)
lasthit = None
lastpick = None
lastcolor = None

def getevent(evt):
    global lasthit, lastpick, lastcolor
    if lasthit != None:
        if lastpick != None: lasthit.modify(lastpick, color=lastcolor)
        else: lasthit.color = vector(lastcolor)
        lasthit = lastpick = None

    hit = scene.mouse.pick
    if hit != None:
        lasthit = hit
        lastpick = None
        if isinstance(hit, curve):  # pick individual point of curve
            lastpick = hit.segment
            lastcolor = hit.point(lastpick)['color']
            hit.modify(lastpick, color=color.red)
        elif isinstance(hit, quad):
            lasthit = hit.v0
            lastcolor = vector(lasthit.color) # make a copy
            lasthit.color = color.red
        else:
            lastcolor = vector(hit.color) # make a copy
            hit.color = color.red

scene.bind("mousedown", getevent)




from vpython import *

# Bruce Sherwood

N = 4 # N by N by N array of atoms
# Surrounding the N**3 atoms is another layer of invisible fixed-position atoms
# that provide stability to the lattice.
k = 1
m = 1
spacing = 1
atom_radius = 0.3*spacing
L0 = spacing-1.8*atom_radius
V0 = pi*(0.5*atom_radius)**2*L0 # initial volume of spring
scene.center = 0.5*(N-1)*vector(1,1,1)
dt = 0.04*(2*pi*sqrt(m/k))
axes = [vector(1,0,0), vector(0,1,0), vector(0,0,1)]

scene.caption= """A model of a solid represented as atoms connected by interatomic bonds.

Right button drag or Ctrl-drag to rotate "camera" to view scene.
To zoom, drag with middle button or Alt/Option depressed, or use scroll wheel.
  On a two-button mouse, middle is left + right.
Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""

class crystal:
        
    def __init__(self,  N, atom_radius, spacing, momentumRange ):
        self.atoms = []
        self.springs = []
        
        # Create (N+2)^3 atoms in a grid; the outermost atoms are fixed and invisible
        for z in range(-1,N+1,1):
            for y in range(-1,N+1,1):
                for x in range(-1,N+1,1):
                    atom = sphere()
                    atom.pos = vector(x,y,z)*spacing
                    atom.radius = atom_radius
                    atom.color = vector(0,0.58,0.69)
                    if 0 <= x < N and 0 <= y < N and 0 <= z < N:
                        p = vec.random()
                        atom.momentum = momentumRange*p
                    else:
                        atom.visible = False
                        atom.momentum = vec(0,0,0)
                    atom.index = len(self.atoms)
                    self.atoms.append( atom )
        for atom in self.atoms:
            if atom.visible:
                if atom.pos.x == 0:
                    self.make_spring(self.atoms[atom.index-1], atom, False)
                    self.make_spring(atom, self.atoms[atom.index+1], True)
                elif atom.pos.x == N-1:
                    self.make_spring(atom, self.atoms[atom.index+1], False)
                else:
                    self.make_spring(atom, self.atoms[atom.index+1], True)

                if atom.pos.y == 0:
                    self.make_spring(self.atoms[atom.index-(N+2)], atom, False)
                    self.make_spring(atom, self.atoms[atom.index+(N+2)], True)
                elif atom.pos.y == N-1:
                    self.make_spring(atom, self.atoms[atom.index+(N+2)], False)
                else:
                    self.make_spring(atom, self.atoms[atom.index+(N+2)], True)
                    
                if atom.pos.z == 0:
                    self.make_spring(self.atoms[atom.index-(N+2)**2], atom, False)
                    self.make_spring(atom, self.atoms[atom.index+(N+2)**2], True)
                elif atom.pos.z == N-1:
                    self.make_spring(atom, self.atoms[atom.index+(N+2)**2], False)
                else:
                    self.make_spring(atom, self.atoms[atom.index+(N+2)**2], True)
    
    # Create a grid of springs linking each atom to the adjacent atoms
    # in each dimension, or to invisible motionless atoms
    def make_spring(self, start, end, visible):
        spring = helix()
        spring.pos = start.pos
        spring.axis = end.pos-start.pos
        spring.visible = visible
        spring.thickness = 0.05
        spring.radius = 0.5*atom_radius
        spring.length = spacing
        spring.start = start
        spring.end = end
        spring.color = color.orange
        self.springs.append(spring)

c = crystal(N, atom_radius, spacing, 0.1*spacing*sqrt(k/m))

while True:
    rate(60)
    for atom in c.atoms:
        if atom.visible:
            atom.pos = atom.pos + atom.momentum/m*dt
    for spring in c.springs:
        spring.axis = spring.end.pos - spring.start.pos
        L = mag(spring.axis)
        spring.axis = spring.axis.norm()
        spring.pos = spring.start.pos+0.5*atom_radius*spring.axis
        Ls = L-atom_radius
        spring.length = Ls
        Fdt = spring.axis * (k*dt * (1-spacing/L))
        if spring.start.visible:
            spring.start.momentum = spring.start.momentum + Fdt
        if spring.end.visible:
            spring.end.momentum = spring.end.momentum - Fdt



from vpython import *
# This program displays most GlowScript 3D objects.

#It also illustrates key features such as mouse handling, rate, and sleep.

# Bruce Sherwood, August 2011; VPython version November 2014

s = sphere(color=color.magenta, radius=0.5, visible=False)
drag = False

def down(ev):
    global drag
    s.pos = scene.mouse.pos
    s.visible = True
    drag = True

def move(ev):
    global drag
    if not drag: return
    s.pos = scene.mouse.pos

def up(ev):
    global drag
    s.visible = False
    drag = False

scene.bind("mousedown", down)

scene.bind("mousemove", move)

scene.bind("mouseup", up)

scene.title = "A display of most GlowScript 3D objects"
scene.width = 640
scene.height = 400
#scene.range = 5
scene.background = color.gray(0.7)
scene.center = vector(0,0.5,0)
scene.forward = vector(-.3,0,-1)
gslabel = label(pos=vector(1.1,2,0), text='GlowScript', 
    xoffset=40, height=16, color=color.yellow)
box(pos=vector(-2,0,0), size=vector(.3,2.5,2.5), color=color.red)
box(pos=vector(.25,-1.4,0), size=vector(4.8,.3,2.5), color=color.red)
cylinder(pos=vector(-2,2,1.25), radius=0.7, axis=vector(0,0,-2.5), color=color.blue)
ball = sphere(pos=vector(2,1,0), radius=0.5, color=color.cyan)
ptr = arrow(pos=vector(0,0,2), axis=vector(2,0,0), color=color.yellow)
cone(pos=vector(-2,0,0), radius=1, length=3, color=color.green, opacity=0.3)
ring(pos=vector(.2,0,0), radius=.6, axis=vector(1,0,0), thickness=0.12, color=color.gray(0.4))
ellipsoid(pos=vector(-.3,2,0), color=color.orange, size=vector(.3,1.5,1.5))
pyramid(pos=vector(.3,2,0), color=vector(0,0.5,.25), size=vector(0.8,1.2,1.2))
spring = helix(pos=vector(2,-1.25,0), radius=0.3, axis=vector(0,1.8,0),
        color=color.orange, thickness=.1)
angle = 0
da = .01

trail = curve(color=color.magenta, radius= .02)
trail.append(vector(1,0,0))
trail.append(vector(1,0,2))
trail.append(vector(2,0,2))

while angle < 3*pi/4:
  rate(100)
  ptr.rotate(angle=da, axis=vector(0,0,1), origin=ptr.pos)
  trail.append(ptr.pos+ptr.axis)
  angle += da

sleep(1) # sleep for 1 second
scene.autoscale = False
scene.append_to_caption("""Drag the mouse and you'll drag a sphere.
On a touch screen, press and hold, then drag.""")

t = 0
dt = .01
y0 = gslabel.pos.y
ball_yo = ball.pos.y
while t < 10:
  rate(1/dt)
  ball.pos.y = ball_yo+0.5*sin(-4*t)
  spring.length = ball.pos.y-spring.pos.y-ball.radius+0.15
  gslabel.yoffset = 28*sin(-4*t)
  t += dt



from setuptools import setup

from distutils.extension import Extension

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
    extensions = cythonize('vpython/cyvector.pyx')
except ImportError:
    extensions = [Extension('vpython.cyvector', ['vpython/cyvector.c'])]


install_requires = ['jupyter', 'jupyter-server-proxy', 'numpy', 'ipykernel',
                    'autobahn>=22.6.1, <27']

setup_args = dict(
    name='vpython',
    packages=['vpython'],
    description='VPython for Jupyter Notebook',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author='John Coady / Ruth Chabay / Bruce Sherwood / Steve Spicklemire',
    author_email='bruce.sherwood@gmail.com',
    url='http://pypi.python.org/pypi/vpython/',
    license='LICENSE.txt',
    keywords='vpython',
    classifiers=[
          'Framework :: IPython',
          'Development Status :: 5 - Production/Stable',
          'Environment :: Web Environment',
          'Intended Audience :: End Users/Desktop',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Topic :: Multimedia :: Graphics :: 3D Modeling',
          'Topic :: Multimedia :: Graphics :: 3D Rendering',
          'Topic :: Scientific/Engineering :: Visualization',
    ],
    ext_modules=extensions,
    install_requires=install_requires,
    python_requires=">=3.7",
    package_data={'vpython': ['vpython_data/*',
                              'vpython_libraries/*',
                              'vpython_libraries/images/*']},
)

try:
    setup(**setup_args)
except SystemExit as e:
    print('Compiled version of vector failed with:')
    print(e)
    print('***** Using pure python version of vector, which is slower.')

    # Do not try to build the extension
    del setup_args['ext_modules']
    setup(**setup_args)



from vpython import *
scene.width = scene.height = 600
# There is an L by L grid of vertex objects, numbered 0 through L-1 by 0 through L-1.
# Only the vertex operators numbered L-2 by L-2 are used to create quads.
# The extra row and extra column of vertex objects simplifies edge calculations.
# The stride length from y = 0 to y = 1 is L.
L = 50
scene.center = vec(0.05*L,0.2*L,0)
scene.range = 1.3*L
## The next line contains LaTeX math notation. See http://www.glowscript.org/docs/VPythonDocs/MathJax.html
#scene.caption = """\\( f(x,y,t) = 0.7+0.2\\sin{(10x)}\\cos{(10y)}\\cos{(2t)} \\)
scene.caption = """<i>f</i>(<i>x,y,t</i>) = 0.7+0.2sin(10<i>x</i>)cos(10<i>y</i>)cos(2<i>t</i>)
<b>Click to toggle between pausing or running.</b>
In GlowScript programs:
   Right button drag or Ctrl-drag to rotate "camera" to view scene.
   To zoom, drag with mid1e button or Alt/Option depressed, or use scroll wheel.
     On a two-button mouse, midd1e is left + right.
   Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""
#MathJax.Hub.Queue(["Typeset",MathJax.Hub]) # format the LaTeX; see http://www.glowscript.org/docs/VPythonDocs/MathJax.html

class plot3D:
    def __init__(self, f, xmin, xmax, ymin, ymax, zmin, zmax):
        # The x axis is labeled y, the z axis is labeled x, and the y axis is labeled z.
        # This is done to mimic fairly standard practive for plotting
        #     the z value of a function of x and y.
        self.f = f
        if not xmin: self.xmin = 0
        else: self.xmin = xmin
        if not xmax: self.xmax = 1
        else: self.xmax = xmax
        if not ymin: self.ymin = 0
        else: self.ymin = ymin
        if not ymax: self.ymax = 1
        else: self.ymax = ymax
        if not zmin: self.zmin = 0
        else: self.zmin = zmin
        if not zmax: self.zmax = 1
        else: self.zmax = zmax
        
        R = L/100
        d = L-2
        xaxis = cylinder(pos=vec(0,0,0), axis=vec(0,0,d), radius=R, color=color.yellow)
        yaxis = cylinder(pos=vec(0,0,0), axis=vec(d,0,0), radius=R, color=color.yellow)
        zaxis = cylinder(pos=vec(0,0,0), axis=vec(0,d,0), radius=R, color=color.yellow)
        k = 1.02
        h = 0.05*L
        text(pos=xaxis.pos+k*xaxis.axis, text='x', height=h, align='center', billboard=True, emissive=True)
        text(pos=yaxis.pos+k*yaxis.axis, text='y', height=h, align='center', billboard=True, emissive=True)
        text(pos=zaxis.pos+k*zaxis.axis, text='z', height=h, align='center', billboard=True, emissive=True)
    
        self.vertices = []
        for x in range(L):
            for y in range(L):
                val = self.evaluate(x,y)
                self.vertices.append(self.make_vertex( x, y, val ))
        
        self.make_quads()
        self.make_normals()
        
    def evaluate(self, x, y):
        d = L-2
        return (d/(self.zmax-self.zmin)) * (self.f(self.xmin+x*(self.xmax-self.xmin)/d, self.ymin+y*(self.ymax-self.ymin)/d)-self.zmin)
    
    def make_quads(self):
        # Create the quad objects, based on the vertex objects already created.
        for x in range(L-2):
            for y in range(L-2):
                v0 = self.get_vertex(x,y)
                v1 = self.get_vertex(x+1,y)
                v2 = self.get_vertex(x+1, y+1)
                v3 = self.get_vertex(x, y+1)
                quad(vs=[v0, v1, v2, v3])
        
    def make_normals(self):
        # Set the normal for each vertex to be perpendicular to the lower left corner of the quad.
        # The vectors a and b point to the right and up around a vertex in the xy plance.
        for i in range(L*L):
            x = int(i/L)
            y = i % L
            if x == L-1 or y == L-1: continue
            v = self.vertices[i]
            a = self.vertices[i+L].pos - v.pos
            b = self.vertices[i+1].pos - v.pos
            v.normal = cross(a,b)
    
    def replot(self):
        for i in range(L*L):
            x = int(i/L)
            y = i % L
            v = self.vertices[i]
            v.pos.y = self.evaluate(x,y)
        self.make_normals()
                
    def make_vertex(self,x,y,value):
        return vertex(pos=vec(y,value,x), color=color.cyan, normal=vec(0,1,0))
        
    def get_vertex(self,x,y):
        return self.vertices[x*L+y]
        
    def get_pos(self,x,y):
        return self.get_vertex(x,y).pos

t = 0
dt = 0.02
def f(x, y):
    # Return the value of the function of x and y:
    return 0.7+0.2*sin(10*x)*cos(10*y)*sin(5*t)

p = plot3D(f, 0, 1, 0, 1, 0, 1) # function, xmin, xmax, ymin, ymax (defaults 0, 1, 0, 1, 0, 1)

run = True
def running(ev):
    global run
    run = not run

scene.bind('mousedown', running)
scene.forward = vec(-0.7,-0.5,-1)

while True:
    rate(1000)
    if run:
        p.replot()
        t += dt



from vpython import *

# Written by Bruce Sherwood, licensed under Creative Commons 4.0.
# All uses permitted, but you must not claim that you wrote it, and
# you must include this license information in any copies you make.
# For details see http://creativecommons.org/licenses/by/4.0

# Gyroscope hanging from a spring

scene.width = scene.height = 600
scene.title = 'Hanging gyroscope'
scene.background = color.white
scene.range = 1.5

top = vector(0,1,0) # where top of spring is held
ks = 100 # spring stiffness
Lspring = 1 # unstretched length of spring
Rspring = 0.03 # radius of spring
Dspring = 0.03 # thickness of spring wire
Lshaft = 1 # length of gyroscope shaft
Rshaft = 0.03 # radius of gyroscope shaft
M = 1 # mass of gyroscope (massless shaft)
Rrotor = 0.4 # radius of gyroscope rotor
Drotor = 0.1 # thickness of gyroscope rotor
Dsquare = 1.4*Drotor # thickness of square that turns with rotor
I = 0.5*M*Rrotor**2 # moment of inertia of gyroscope
omega = 40 # angular velocity of rotor along axis
g = 9.8
Fgrav = vector(0,-M*g,0)
precession = M*g*(Lshaft/2)/(I*abs(omega)) # exact precession angular velocity
phi = atan(precession**2*(Lshaft/2)/g) # approximate angle of spring to vertical
s = M*g/(ks*cos(phi)) # approximate stretch of spring
# Refine estimate of angle of spring to vertical:
phi = 1/( ((I*abs(omega))/(M*Lshaft/2))**2/(g*Lshaft/2)-(Lspring+s)/(Lshaft/2) )
# Refine again:
s = M*g/(ks*cos(phi))
phi = 1/( ((I*abs(omega))/(M*Lshaft/2))**2/(g*Lshaft/2.)-(Lspring+s)/(Lshaft/2) )

support = box(pos=top+vector(0,0.01,0), size=vector(0.2,0.02,0.2), color=color.green)
spring = helix(pos=top, axis=vector(-(Lspring+s)*sin(phi),-(Lspring+s)*cos(phi),0), coils=10,
               radius=Rspring, thickness=Dspring, color=vector(1,0.7,0.2))
a = vector(1,0,0)
shaft = cylinder(axis=Lshaft*a, radius=Rshaft, color=vector(0.85,0.85,0.85))
rotor = cylinder(pos=0.5*a*(Lshaft-Drotor), opacity=0.2,
                 axis=Drotor*a, radius=Rrotor, color=vector(0.5,0.5,0.5))
stripe = box(pos=rotor.pos+0.5*vector(Drotor,0,0),
              size=vector(0.03*Rrotor,2*Rrotor,0.03*Rrotor), color=color.black)
gyro = compound([rotor, shaft, stripe])
gyro.rotate(axis=vector(0,1,0), angle=-pi)
gyro.pos = top+spring.axis+0.5*Lshaft*norm(gyro.axis)

#cm = gyro.pos+0.5*Lshaft*gyro.axis # center of mass of shaft
Lrot = I*omega*gyro.axis
dt = 0.005

Lrotarrow = arrow(length=0, shaftwidth=Rshaft, color=color.red, visible=False)
Lrotscale = 0.07
torquearrow = arrow(length=0, shaftwidth=Lrotarrow.shaftwidth, color=color.cyan, visible=False)
torquescale = 0.11
visLrot = False
visTorque = False

def Runb(r):
    global run
    run = not run
    if run:
        r.text = "Pause"
    else:
        r.text = "Run"

Runbutton = button(text="Run", bind=Runb)

def ResetRunbutton():
    global run
    run = False
    Runbutton.text = "Run"

scene.append_to_caption('   ')

def SPINbutton(s):
    global omega
    omega = -omega
    reset()

button(text='Flip spin direction', bind=SPINbutton)
scene.append_to_caption("   ")

def Lbutton(b):
    global visLrot
    visLrot = not visLrot

checkbox(bind=Lbutton)
scene.append_to_caption(" Show Lrot (red)   ")

def Tbutton(b):
    global visTorque
    visTorque = not visTorque

checkbox( bind=Tbutton)
scene.append_to_caption(" Show torque about CM (cyan)")

scene.append_to_caption("\n\nChoose initial conditions:  ")

# Create a drop-down menu (a "select" object). Set up the options to appear:
menu_options = []
menu_options.append("Pure precession")
menu_options.append("More general motion")

def handlemenu(m): # come here when a change is made in the menu choice
    global pureprecession
    val = m.index
    pureprecession = (val == 0)
    reset()

menu(choices=menu_options, selected=menu_options[0], bind=handlemenu)

pureprecession = True
initial = []

run = False

def reset(first=False):
    global p, pureprecession, t, initial, Lrot
    if first: # take advantage of the existing initializations at the start of the program
        # vector(vector) to make nonmutable copies of the initial vslues
        initial = [vector(gyro.pos), vector(gyro.axis), vector(spring.axis), vector(Lrot),\
                   vector(Lrotarrow.pos), vector(Lrotarrow.axis),\
                   vector(torquearrow.pos), vector(torquearrow.axis)]
    else:
        gyro.pos = initial[0]
        gyro.axis = initial[1]
        spring.axis = initial[2]
        Lrot = initial[3]
        Lrotarrow.pos = initial[4]
        Lrotarrow.axis = initial[5]
        torquearrow.pos = initial[6]
        torquearrow.axis = initial[7]
    t = 0
    if pureprecession:
        p = vector(0, 0, M*precession*(Lshaft/2+(Lspring+s)*sin(phi)))
    else:
        p = vector(0,-1,M*precession*(Lshaft/2+(Lspring+s)*sin(phi)))
    if omega < 0:
        p = -p
        Lrot = -Lrot
    Lrotarrow.visible = False
    torquearrow.visible = False
    ResetRunbutton()

reset(True)

while True:
    rate(100)
    if not run: continue
    
    Fspring = -ks*norm(spring.axis)*(mag(spring.axis)-Lspring)
    # Calculate torque about center of mass:
    torque = cross(-0.5*Lshaft*gyro.axis,Fspring)
    Lrot = Lrot + torque*dt
    p = p + (Fgrav+Fspring)*dt
    gyro.pos = gyro.pos+(p/M)*dt

    # Update positions of shaft, rotor, spring, stripes
    if omega > 0:
        gyro.axis = norm(Lrot)
    else:
        gyro.axis = -norm(Lrot)
    # shaft rotated, adjust connection to spring
    #gyro.pos = cm-0.5*Lshaft*gyro.axis 
    spring.axis = gyro.pos-0.5*Lshaft*norm(gyro.axis) - top
    # spin easier to see if slower than actual omega
    gyro.rotate(angle=omega*dt/4)

    # Update arrows representing angular momentum and angular impulse
    Lrotarrow.visible = visLrot
    torquearrow.visible = visTorque
    if omega > 0:
        Lrotarrow.pos = gyro.pos + 0.5*Lshaft*gyro.axis
    else:
        Lrotarrow.pos = gyro.pos - 0.5*Lshaft*gyro.axis
    Lrotarrow.axis = Lrot*Lrotscale
    torquearrow.pos = gyro.pos
    torquearrow.axis = torque*torquescale
    t = t+dt



import time

from vpython import *
scene.visible = False # while preparing the scene
import time

# Bruce Sherwood

# A surreal scene that illustrates many of the features of GlowScript

# Add instructions below the display
s = "<b>Fly through the scene:</b><br>"
s += "    drag the mouse or your finger above or below the center of the scene to move forward or backward;<br>"
s += "    drag the mouse or your finger right or left to turn your direction of motion.<br>"
s += "(Normal GlowScript rotate and zoom are turned off in this program.)"
scene.caption = s

ycenter = 2
scene.width = 800
scene.height = 400
scene.range = 12
scene.center = vector(0,ycenter,0)
##scene.userspin = False
##scene.userzoom = False
scene.background = color.gray(0.5)
scene.ambient = color.gray(0.4)

def hourminute():
    #GlowScript JavaScript date machinery
    #d = Date()
    #hour = d.getHours() % 12
    #minute = d.getMinutes()
    now = time.localtime(time.time())
    hour = now[3] % 12
    minute = now[4]
    return (hour, minute)

class analog_clock:
    def __init__(self, pos, radius, axis):
        self.pos = pos
        self.axis = axis
        self.radius = radius
        self.spheres = []
        self.hour = 0
        self.minute = -1
        for n in range(12):
            sp = sphere(pos=pos+(radius*scene.up).rotate(angle=-2*pi*n/12, axis=axis),
                    size=(2*radius/20)*vector(1,1,1),
                    color=color.hsv_to_rgb(vector(n/12,1,1)) )
            self.spheres.append(sp)
        self.hand = arrow(pos=self.pos, axis=0.95*radius*scene.up,
                    shaftwidth=radius/10, color=color.cyan)
        self.update()
        
    def update(self):
        hour, minute = hourminute()
        hour = hour % 12
        if self.hour == hour and self.minute == minute: return
        self.hand.axis = (0.95*self.radius*scene.up).rotate(
                    axis=vector(0,0,1), angle=-2*pi*minute/60)
        self.spheres[self.hour].size = (2*self.radius/20)*vector(1,1,1)
        self.spheres[hour].size = (2*self.radius/10)*vector(1,1,1)
        self.hour = hour
        self.minute = minute

grey = color.gray(0.8)
Nslabs = 8
R = 10
w = 5
d = 0.5
h = 5
photocenter = 0.15*w

# The floor, central post, and ball atop the post
floor = box(pos=vector(0,-0.1,0),size=vector(.2,24,24), axis=vector(0,1,0), texture=textures.wood)
pole= cylinder(pos=vector(0,0,0),axis=vector(0,1,0), size=vector(h,0.4,0.4), color=color.red)
sphere(pos=vector(0,h,0), size=vector(1,1,1), color=color.red)

# Set up the gray slabs, including a portal
for i in range(Nslabs):
    theta = i*2*pi/Nslabs
    c = cos(theta)
    s = sin(theta)
    xc = R*c
    zc = R*s
    if i == 2: # Make a portal
        box(pos=vector(-3.*w/8.,0.75*h/2.,R),
            size=vector(0.5*w/2,0.75*h,d), color=grey)
        box(pos=vector(3.*w/8.0,0.75*h/2.,R),
            size=vector(0.5*w/2,0.75*h,d), color=grey)
        box(pos=vector(0,0.85*h,R),
            size=vector(w,0.3*h,d), color=grey)
    else:
        slab = box(pos=vector(R*c, h/2., R*s), axis=vector(c,0,s),
                   size=vector(d,h,w), color=grey)
        if i != 6:
            T = textures.flower
            if (i == 7 or i == 4): T = textures.rug
            box(pos=slab.pos,
                size=vec(1.1*d,0.9*4*photocenter,0.9*4*photocenter), axis=vec(c,0,s),
                    texture=T)

entry = text(pos=vector(0,4.4,R+d/2), text='Surreal\nStonehenge', align='center',
                depth=0.3, height=0.5)
                
B = text(pos=vector(0.4*R,0,-1*R), text='B', height=2, align='center', font='serif',
                color=color.magenta, depth=1)

gh = 1
ga = 1
gr = 0.05
rgear = 0.7
tgear = ga/5 # gear thickness
support = extrusion(pos=vector(7,gh/2,10), path=[vector(0,0,0), vector(0,gh,0), vector(0,gh,ga), vector(0,0,ga)],
            shape=shapes.circle(radius=gr), color=vector(1,.7,0))
support.clone(pos=vector(7+2*rgear,gh/2,10))
gear1 = extrusion(pos=support.pos+vector(0,gh/2,0), path=[vector(0,0,-tgear/2), vector(0,0,tgear/2)],
            shape=shapes.gear(radius=rgear), color=color.gray(0.85), texture=textures.metal)
gear2 = gear1.clone(pos=gear1.pos+vector(2*rgear,0,0))
gear1.rotate(angle=-0.4*2*pi/20, axis=vector(0,0,1)) # default is 20 teeth

# Decorate back slab with a wood box and a clock
box(pos=vector(0,h/2.,-R+d/2+0.1), size=vector(w/2.,w/2.,0.2), texture=textures.wood)
clock = analog_clock(vector(0,h/2.,-R+d/2+0.2+0.2*h/10), 0.2*w, vector(0,0,1))

# Draw guy wires from the top of the central post
Nwires = 32
for i in range(Nwires):
    theta = i*2*pi/Nwires
    L = vector(R * cos(theta), -h - 0.1, R * sin(theta))
    cylinder(pos=vector(0,h,0), axis=L, size=vector(mag(L),.04,.04), color=vector(1,0.7,0))

# Display a pyramid
pyramid(pos=vector(-4,0,-5), size=vector(2,2,2), axis=vector(0,3,0), color=vector(0,.5,0), texture=textures.rough)

# Display smoke rings rising out of a black tube
smoke = []
Nrings = 20
x0, y0, z0 = -5, 1.5, -2
r0 = 0.075
spacing = 0.2
thick = r0/3
dr = 0.0075
dthick = thick/Nrings
gray = 1
cylinder(pos=vector(x0,0,z0), axis=vector(0,y0+r0,0), radius=1.5*(r0+thick), color=color.black)

# Create the smoke rings
for i in range(Nrings):
  smoke.append(ring(pos=vector(x0,y0+spacing*i,z0), axis=vector(0,1,0),
                radius=r0+dr*i, thickness=thick-dthick*i))
y = 0
dy = spacing/20
top = Nrings-1

# Log rolls back and forth between two stops
rlog = 1
wide = 4
zpos = 2
zface = 5
tlogend = 0.2
v0 = 0.3
v = v0
omega = -v0 / rlog
theta = 0
dt = 0.1
tstop = 0.3
logcyl = cylinder(pos=vector(-wide, rlog, zpos), size=vector(zface - zpos, 2, 2),
    axis=vector(0, 0, 1), texture=textures.granite)
leftstop = box(pos=vector(-wide-rlog-tstop/2,0.6*rlog,(zpos+zface)/2),
    size=vector(tstop, 1.2*rlog, (zface-zpos)), color=color.red, emissive=True)
rightstop = box(pos=vector(wide+rlog+tstop/2,0.6*rlog,(zpos+zface)/2),
    size=vector(tstop, 1.2*rlog, (zface-zpos)), color=color.red, emissive=True)

# Run a ball up and down the pole
y1 = 0.2*h
y2 = 0.7*h
rball = 0.4
Dband = 1.3 * pole.size.y
cylinder(pos=vector(0,y1-0.9*rball,0), axis=vector(0,1,0), size=vector(0.1,Dband,Dband), color=color.green)
cylinder(pos=vector(0,y2+0.9*rball,0), axis=vector(0,1,0), size=vector(0.1,Dband,Dband), color=color.green)
vball0 = 0.3*v0
vball = vball0
ballangle = 0.05*pi
ball = []
ball.append(sphere(pos=vector(0,0,0), size=2*rball*vector(1,1,1), color=color.blue))
for nn in range(4):
    cc = cone(pos=vector(0,0,0)+vector(0.8*rball,0,0), axis=vector(3*rball,0,0),
              size=rball*vector(3,1,1), color=color.yellow)
    cc.rotate(angle=0.5*nn*pi, axis=vector(0,1,0), origin=vector(0,0,0))
    ball.append(cc)
ball = compound(ball)
ball.pos = vector(0,y1,0)

# A table with a mass-spring object sliding on it
table = cone(pos=vector(0.4*R, h/4, -.3*R), size=vector(h/4, 0.6 * R, 0.6 * R), 
  axis=vector(0, -1, 0), texture={'file':textures.wood_old, 'turn':1})
tabletop = table.pos
rspring = 0.02 * h
Lspring = .15 * R
Lspring0 = .1 * R
hmass = 4 * rspring
post = cylinder(pos=tabletop, axis=vector(0, 1, 0), size=vector(2 * hmass, .4, .4), color=color.gray(.6))
spring = helix(pos=post.pos + vector(0, hmass/2, 0), size=vector(Lspring, 2 * rspring, 2 * rspring),
               color=color.orange, thickness=rspring)
mass = cylinder(pos=post.pos + vector(Lspring, 0, 0), axis=vector(0, 1, 0),
                size=vector(hmass, .04 * R, .04 * R), color=color.orange)
mass.p = vector(10, 0, 5)
mass.m = 1
kspring = 200
deltat = .01

# Display an ellipsoid
Rcloud = 0.8*R
omegacloud =3*v0/Rcloud
cloud = sphere(pos=vector(0,0.7*h,-Rcloud), size=vector(5,2,2),
                  color=color.green, opacity=0.3)

rhairs = 0.025 # half-length of crosshairs
dhairs = 2 # how far away the crosshairs are
maxcosine = dhairs/sqrt(rhairs**2+dhairs**2) # if ray inside crosshairs, don't move
haircolor = color.black
roam = 0

scene.waitfor("textures")
scene.visible = True # finished preparing the scene

roam = False

def setroam(evt):
    global roam
    roam = not roam

scene.bind("mousedown mouseup", setroam)

hue = 0
dhue = 0.01
gangle = 0.03 # incremental rotation of the gears

while True:
    rate(30)

    # If in roaming mode, change center and forward according to mouse position
    if roam:
        ray = scene.mouse.ray
        if abs(ray.dot(scene.forward)) < maxcosine: # do something only if outside crosshairs
            newray = norm(vector(ray.x, 0, ray.z))
            angle = asin(scene.forward.cross(newray).dot(scene.up))
            scene.camera.rotate(angle=angle/30, axis=scene.up)
            scene.camera.pos = scene.camera.pos + (ray.y/2)*norm(scene.camera.axis)
    
    hue += dhue
    entry.color = color.hsv_to_rgb(vector(hue,1,1))
    
    B.rotate(angle=0.05, axis=scene.up)
    
    gear1.rotate(angle=gangle, axis=vector(0,0,1))
    gear2.rotate(angle=-gangle, axis=vector(0,0,1))

    # Roll the log
    theta = theta + omega*dt
    logcyl.pos.x = logcyl.pos.x+v*dt
    logcyl.rotate(angle=omega*dt, axis=vector(0,0,1))
    if logcyl.pos.x >= wide:
        v = -v0
        omega = -v/rlog
        if rightstop.color.equals(color.red):
            rightstop.color = color.cyan
        else:
            rightstop.color = color.red
    if logcyl.pos.x <= -wide:
        v = +v0
        omega = -v/rlog
        if leftstop.color.equals(color.red):
            leftstop.color = color.cyan
        else:
            leftstop.color = color.red

    # Move the cloud
    cloud.rotate(angle=omegacloud*dt, origin=vector(0,0,0), axis=vector(0,1,0))

    # Run the ball up and down
    ball.pos.y = ball.pos.y+vball*dt
    ball.rotate(angle=ballangle, axis=vector(0,1,0))
    if ball.pos.y >= y2:
        vball = -vball0
        ballangle = -ballangle
    if ball.pos.y <= y1:
        vball = +vball0
        ballangle = -ballangle

    # Move the smoke rings
    for i in range(Nrings):
        # Raise the smoke rings
        smoke[i].pos = smoke[i].pos+vector(0,dy,0)
        smoke[i].radius = smoke[i].radius+(dr/spacing)*dy
        smoke[i].thickness = smoke[i].thickness - (dthick/spacing)*dy
    y = y+dy
    if y >= spacing:
        # Move top ring to the bottom
        y = 0
        smoke[top].pos = vector(x0, y0, z0)
        smoke[top].radius = r0
        smoke[top].thickness = thick
        top = top-1
    if top < 0:
        top = Nrings-1
        
    # Update the mass-spring motion
    F = -kspring * (spring.size.x - Lspring0) * spring.axis.norm()
    mass.p = mass.p + F * deltat
    mass.pos = mass.pos + (mass.p / mass.m) * deltat
    spring.axis = mass.pos + vector(0, hmass / 2, 0) - spring.pos

    # Update the analog clock on the back slab
    clock.update()



from vpython import *

# Convert 3D .stl file ("stereo lithography") to VPython 7 object.

# Limitations:
#    Code for binary files needs to be updated to VPython 7.
#    Does not deal with color.
#    Does not assign texpos values to vertex objects,
#      so cannot add a meaningful texture to the final compound object.

# Original converter and STLbot by Derek Lura 10/06/09
# Be sure to look at the bottom of the STLbot figure!
# Part1.stl found at 3Dcontentcentral.com; also see 3dvia.com

# Factory function and handling of binary files by Bruce Sherwood 1/26/10
# Conversion to VPython 7 by Bruce Sherwood 2018 May 8

# Give this factory function an .stl file and it returns a compound object,
# to permit moving and rotating.

# Specify the file as a file name.

# See http://en.wikipedia.org/wiki/STL_(file_format)
# Text .stl file starts with a header that begins with the word "solid".
# Binary .stl file starts with a header that should NOT begin with the word "solid",
# but this rule seems not always to be obeyed.
# Currently the 16-bit unsigned integer found after each triangle in a binary
# file is ignored; some versions of .stl files put color information in this value.

def stl_to_triangles(fileinfo): # specify file
    # Accept a file name or a file descriptor; make sure mode is 'rb' (read binary)
    fd = open(fileinfo, mode='rb')
    text = fd.read()
    tris = [] # list of triangles to compound
    keywords = [b'outer', b'endloop', b'endfacet', b'solid', b'endsolid']
    if False: # prevent executing code for binary file
        pass
    # The following code for binary files must be updated:
#    if chr(0) in text: # if binary file
#        text = text[84:]
#        L = len(text)
#        N = 2*(L//25) # 25/2 floats per point: 4*3 float32's + 1 uint16
#        triPos = []
#        triNor = []
#        n = i = 0
#        while n < L:
#            triNor[i] = fromstring(text[n:n+12], float32)
#            triPos[i] = fromstring(text[n+12:n+24], float32)
#            triPos[i+1] = fromstring(text[n+24:n+36], float32)
#            triPos[i+2] = fromstring(text[n+36:n+48], float32)
#            colors = fromstring(text[n+48:n+50], uint16)
#            if colors != 0:
#                print '%x' % colors
#            if triNor[i].any():
#                triNor[i] = triNor[i+1] = triNor[i+2] = norm(vector(triNor[i]))
#            else:
#                triNor[i] = triNor[i+1] = triNor[i+2] = \
#                    norm(cross(triPos[i+1]-triPos[i],triPos[i+2]-triPos[i]))
#            n += 50
#            i += 3
    else:
        fd.seek(0)
        fList = fd.readlines()
        print('Number of lines =', len(fList))
    
        # Decompose list into vertex positions and normals
        ret = [] # will return a list of compounds if necessary
        vs = []
        vertices = 0
        for line in fList:
            FileLine = line.split( )
            first = FileLine[0]
            if first == b'facet':
                N = vec(float(FileLine[2]), float(FileLine[3]), float(FileLine[4]))
            elif first == b'vertex':
                vertices += 1
                vs.append( vertex(pos=vec(float(FileLine[1]), float(FileLine[2]), float(FileLine[3])), normal=N, color=color.white) )
                if len(vs) == 3:
                    tris.append(triangle(vs=vs))
                    vs = []
                    if vertices > 64000:
                        print(vertices)
                        ret.append(compound(tris))
                        tris = []
                        vertices = 0
            elif first in keywords:
                pass
            else:
                print(line) # for debugging
    if len(tris) > 0: ret.append(compound(tris))
    if len(ret) == 1: return ret[0]               
    else: return ret

if __name__ == '__main__':
    man = stl_to_triangles('STLbot.stl')
    man.pos = vec(-200,0,0)
    man.color = color.cyan
    part = stl_to_triangles('Part1.stl')
    part.size *= 200
    part.pos = vec(250,0,0)
    part.color = color.orange
    print('Done')
